# Copyright (c) 2023 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import json
import re
import typing

from oslo_log import log
import pytest
import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import octavia
from tobiko.openstack import stacks
from tobiko.openstack import topology
from tobiko import podified
from tobiko.shell import sh
from tobiko import tripleo

LOG = log.getLogger(__name__)


class BaseSecurityGroupTest(testtools.TestCase):

    _ovn_nb_db = None
    _host_ssh_client = None
    _container_runtime_name = None
    _container_name = None

    def setUp(self):
        super(BaseSecurityGroupTest, self).setUp()
        self.ovn_controller_agents = neutron.list_networking_agents(
            binary=neutron.OVN_CONTROLLER)
        if len(self.ovn_controller_agents) < 1:
            self.skip(f"No running {neutron.OVN_CONTROLLER} agents found. "
                      f"Stateless Security Group tests requires ML2/OVN "
                      f"deployment.")

    @property
    def ovn_nb_db(self):

        def get_podified_ovn_nb_db():
            nb_db = podified.get_ovndbcluter(
                'ovndbcluster-nb')['status']['dbAddress']
            ssl_params = ''
            if 'ssl' in nb_db:
                # SSL options obtained from the container under test
                command = ""
                if topology.get_openstack_topology().has_containers:
                    command += (f"{self.container_runtime_name} exec "
                                f"{self.container_name} ")
                command += "ps -o command -C ovn-controller --no-headers -ww"
                command_result = sh.execute(command,
                                            ssh_client=self.host_ssh_client,
                                            sudo=True).stdout.strip()
                for param in ('p', 'c', 'C'):
                    # the matched strings start with a space
                    ssl_params += re.search(r' -{} [^\s]+'.format(param),
                                            command_result).group()
            return nb_db + ssl_params

        def get_ovn_nb_db():
            command_result = sh.execute(
                "ovs-vsctl get open . external_ids:ovn-remote | "
                "sed -e 's/\"//g' | sed 's/6642/6641/g'",
                ssh_client=self.host_ssh_client,
                sudo=True)
            if 'ovsdbserver-sb' in command_result.stdout:
                nb_db = command_result.stdout.replace('ovsdbserver-sb',
                                                      'ovsdbserver-nb')
            else:
                nb_db = command_result.stdout
            ssl_params = ''
            if 'ssl' in command_result.stdout:
                ssl_params = ' -p {} -c {} -C {} '.format(
                    '/etc/pki/tls/private/ovn_controller.key',
                    '/etc/pki/tls/certs/ovn_controller.crt',
                    '/etc/ipa/ca.crt')
            return nb_db + ssl_params

        if not self._ovn_nb_db:
            if podified.has_podified_cp():
                self._ovn_nb_db = get_podified_ovn_nb_db()
            else:
                self._ovn_nb_db = get_ovn_nb_db()
        return self._ovn_nb_db

    @property
    def host_ssh_client(self):
        if not self._host_ssh_client:
            for ovn_controller_agent in self.ovn_controller_agents:
                candidate_ssh_client = topology.get_openstack_node(
                        hostname=ovn_controller_agent['host']).ssh_client
                # some of the ovn-controller hosts may be not ssh'able
                # (specifically, this happens with Podified Controlplane hosts)
                if candidate_ssh_client is not None:
                    self._host_ssh_client = candidate_ssh_client
                    break
        return self._host_ssh_client

    @property
    def container_runtime_name(self):
        if not self._container_runtime_name:
            if tripleo.has_overcloud():
                self._container_runtime_name = (
                    tripleo.get_container_runtime_name())
            elif podified.has_podified_cp():
                self._container_runtime_name = (
                    podified.get_container_runtime_name())
            else:
                self._container_runtime_name = 'docker'
        return self._container_runtime_name

    @property
    def container_name(self):
        if self._container_name is None:
            os_topology = topology.get_openstack_topology()
            if os_topology.has_containers:
                self._container_name = topology.get_agent_container_name(
                    neutron.OVN_CONTROLLER
                )
            else:
                self._container_name = ""
        return self._container_name

    def _check_sg_rule_in_ovn_nb_db(self, rule_id: str, expected_action: str):
        os_topology = topology.get_openstack_topology()
        command = ""
        if os_topology.has_containers:
            command += (
                f"{self.container_runtime_name} exec {self.container_name} ")
        command += (
            f"ovn-nbctl --format json --no-leader-only --db={self.ovn_nb_db} "
            f"find ACL external_ids:\"neutron\\:security_group_rule_id\"="
            f"\"{rule_id}\""
        )
        command_result = sh.execute(
            command, ssh_client=self.host_ssh_client, sudo=True)
        acl_rule = json.loads(command_result.stdout)
        self._assert_acl_action(acl_rule, expected_action)

    def _assert_acl_action(self, acl: dict, expected_action: str):
        action_column = acl['headings'].index('action')
        self.assertEqual(
            expected_action,
            acl['data'][0][action_column])

    def _check_sg_rules_in_ovn_nb_db(self, sg: dict, expected_action: str):
        sg_rules_ids = [
            rule['id'] for rule in sg['security_group_rules']]
        for sg_rule_id in sg_rules_ids:
            self._check_sg_rule_in_ovn_nb_db(sg_rule_id, expected_action)


@pytest.mark.skip_during_ovn_migration
@neutron.skip_if_missing_networking_extensions('stateful-security-group')
class StatelessSecurityGroupTest(BaseSecurityGroupTest):

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(
        stacks.CirrosServerWithDefaultSecurityGroupStackFixture)

    def test_default_security_group_is_stateful(self):
        """Test that default security group is always stateful.

        This test checks if default SG created for the project is stateful
        and if OVN's ACLs corresponding to the SG's rules have correct
        action which is "allow-related".

        Steps:
        1. Get default SG for the project (it's always created automatically),
        2. Check if ACLs corresponding to the rules from that SG have
           "action-related" action,
        3. Add new SG rule in the default SG,
        4. Check action of the ACL corresponding to the newly created SG rule.
        """

        default_sg = neutron.get_default_security_group(
            project_id=self.stack.project)
        self.assertTrue(default_sg['stateful'])
        self._check_sg_rules_in_ovn_nb_db(default_sg,
                                          neutron.STATEFUL_OVN_ACTION)
        new_rule = neutron.create_security_group_rule(
            default_sg['id'],
            port_range_min=1111,
            port_range_max=1111,
            ethertype="IPv4",
            protocol="tcp",
            description="test_default_security_group_is_stateful rule",
            direction="ingress"
        )
        self._check_sg_rule_in_ovn_nb_db(new_rule['id'],
                                         neutron.STATEFUL_OVN_ACTION)

    def test_security_group_stateful_to_stateless_switch(self):
        """Test that security group can be switched from stateful to stateless.

        This test initially checks if newly created SG is stateful by default
        and if OVN's ACLs corresponding to the SG's rules have correct
        action which is "allow-related".
        Later it also checks if SG can be updated to be stateless and if OVN's
        ACLs corresponding to the SG's rules are properly updated too.

        Steps:
        1. Create SG for the project,
        2. Check if ACLs corresponding to the rules from that SG have
           "action-related" action,
        3. Add new SG rule in the SG,
        4. Check action of the ACL corresponding to the newly created SG rule,
        5. Update SG to be stateless,
        6. Check if ACLs corresponding to the rules from that SG have
           "action-stateless" action,
        7. Add new SG rule in the SG,
        8. Check action of the ACL corresponding to the newly created SG rule,
        9. Update SG to be stateful again,
        10. Add new SG rule in the SG,
        11. Check action of the ACL corresponding to the newly created SG rule,
        """
        sg = neutron.create_security_group(
            name="test_new_security_group_is_statefull_SG",
        )
        self.assertTrue(sg['stateful'])
        self._check_sg_rules_in_ovn_nb_db(sg, neutron.STATEFUL_OVN_ACTION)
        new_rule = neutron.create_security_group_rule(
            sg['id'],
            port_range_min=1111,
            port_range_max=1111,
            ethertype="IPv4",
            protocol="tcp",
            description="stateful SG rule 1",
            direction="ingress"
        )
        self._check_sg_rule_in_ovn_nb_db(new_rule['id'],
                                         neutron.STATEFUL_OVN_ACTION)

        # Update to stateless
        neutron.update_security_group(sg['id'], stateful=False)
        sg = neutron.get_security_group(sg['id'])
        self.assertFalse(sg['stateful'])
        self._check_sg_rules_in_ovn_nb_db(sg, neutron.STATELESS_OVN_ACTION)
        new_rule = neutron.create_security_group_rule(
            sg['id'],
            port_range_min=2222,
            port_range_max=2222,
            ethertype="IPv4",
            protocol="tcp",
            description="stateless SG rule",
            direction="ingress"
        )
        self._check_sg_rule_in_ovn_nb_db(new_rule['id'],
                                         neutron.STATELESS_OVN_ACTION)

        # And get back to stateful
        neutron.update_security_group(sg['id'], stateful=True)
        sg = neutron.get_security_group(sg['id'])
        self.assertTrue(sg['stateful'])
        self._check_sg_rules_in_ovn_nb_db(sg, neutron.STATEFUL_OVN_ACTION)
        new_rule = neutron.create_security_group_rule(
            sg['id'],
            port_range_min=3333,
            port_range_max=3333,
            ethertype="IPv4",
            protocol="tcp",
            description="stateful SG rule 2",
            direction="ingress"
        )
        self._check_sg_rule_in_ovn_nb_db(new_rule['id'],
                                         neutron.STATEFUL_OVN_ACTION)

    def test_create_stateless_security_group(self):
        """Test that stateless security group can be created.

        This test checks if creation of the stateless SG is working fine
        and if OVN's ACLs corresponding to the SG's rules have correct
        action which is "allow-stateless".

        Steps:
        1. Create stateless security group,
        2. Check if ACLs corresponding to the rules from that SG have
           "action-stateless" action,
        3. Add new SG rule in the SG,
        4. Check action of the ACL corresponding to the newly created SG rule.
        """
        sg = neutron.create_security_group(
            name="test_stateless_SG",
            stateful=False
        )
        self.assertFalse(sg['stateful'])
        self._check_sg_rules_in_ovn_nb_db(sg, neutron.STATELESS_OVN_ACTION)
        new_rule = neutron.create_security_group_rule(
            sg['id'],
            port_range_min=1111,
            port_range_max=1111,
            ethertype="IPv4",
            protocol="tcp",
            description="test_new_security_group_is_statefull_SG rule",
            direction="ingress"
        )
        self._check_sg_rule_in_ovn_nb_db(new_rule['id'],
                                         neutron.STATELESS_OVN_ACTION)


@neutron.skip_if_missing_networking_extensions('port-security',
                                               'stateful-security-group')
class CirrosServerWithStatelessSecurityGroupFixture(
        stacks.CirrosServerStackFixture):
    """Heat stack for testing a floating IP instance with port security"""

    #: Resources stack with security group to allow ping Nova servers
    security_groups_stack: tobiko.RequiredFixture = \
        tobiko.required_fixture(stacks.StatelessSecurityGroupFixture)

    @property
    def security_groups(self) -> typing.List[str]:
        """List with Stateless security group"""
        return [self.security_groups_stack.security_group_id]


@pytest.mark.skip_during_ovn_migration
@neutron.skip_if_missing_networking_extensions('stateful-security-group')
class StatelessSecurityGroupInstanceTest(BaseSecurityGroupTest):

    #: Resources stack with Nova server to send messages to
    vm = tobiko.required_fixture(
        CirrosServerWithStatelessSecurityGroupFixture)

    lb_member_server = tobiko.required_fixture(
        stacks.OctaviaServerStackFixture)

    def test_no_conntrack_entries_related_to_stateless_sg(self):
        """ Test that there is no conntrack entry related to stateless SG.

        This test ensures that there is no conntrack entry for connection
        that passes stateless security group.

        Steps:
        1. Create server with stateless SG,
        2. Allow SSH to that instance
        3. Make SSH connection to the instance,
        4. Ensure on compute node that there are no conntrack entries
           related to that connection there,
        """
        host_ssh_client = topology.get_openstack_node(
            hostname=self.vm.hypervisor_hostname).ssh_client
        vm_ip_address = self.vm.find_fixed_ip(ip_version=4)

        # Now lets make ssh connection to the vm and then check in conntrack if
        # entry is there or not (it shouldn't be)
        sh.execute('hostname', ssh_client=self.vm.ssh_client)
        conntrack_list_result = sh.execute(
            "conntrack -L --proto tcp --dport 22 --dst %s" % vm_ip_address,
            ssh_client=host_ssh_client, sudo=True)
        # And ensure that there is no entry found in conntrack
        self.assertEqual("", conntrack_list_result.stdout)
        self.assertTrue(
            "0 flow entries have been shown" in conntrack_list_result.stderr)

    @neutron.skip_unless_is_ovn()
    @keystone.skip_if_missing_service(name='octavia')
    def test_sg_and_loadbalancer_together(self):
        """Test that stateless SG will work fine together with Octavia LB.

        This test ensures that stateless SG works fine even when Octavia
        Loadbalancer is connected to the same network as VM using stateless SG.
        This test is coverage for the issue found in the past and reported in
        https://bugs.launchpad.net/neutron/+bug/2024502

        Steps:
        1. Create LB, listener, pool and member connected to that pool,
        2. Create VM connected to the same network as pool's member. This VM
           uses Stateless SG,
        3. Check SSH connectivity to the VM created in step 2.
        """

        octavia.deploy_ipv4_ovn_lb(
            servers_stacks=[self.lb_member_server]
        )
        # Ensure that ssh connection to the vm works as expected
        sh.execute('hostname', ssh_client=self.vm.ssh_client)



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from datetime import datetime
import math
import random
import re
import socket
import time
import urllib.parse

import netaddr
from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import glance
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import stacks
from tobiko.openstack import tests
from tobiko.openstack import topology
from tobiko.tests.faults.ha import test_cloud_recovery
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.tripleo import containers
from tobiko.tripleo import nova
from tobiko.tripleo import pacemaker
from tobiko.tripleo import topology as tripleo_topology
from tobiko import tripleo


CONF = config.CONF
LOG = log.getLogger(__name__)

network_disruption = """
 sudo iptables-save > ~/working.iptables.rules &&
 sudo iptables -I INPUT 1 -m state --state RELATED,ESTABLISHED -j ACCEPT &&
 sudo iptables -I INPUT 2 -p tcp -m state --state NEW -m tcp --dport 22 -j \
 ACCEPT &&
 sudo iptables -I INPUT 3 ! -i lo -j DROP &&
 sudo iptables -I OUTPUT 1 -p tcp --sport 22 -j ACCEPT &&
 sudo iptables -I OUTPUT 2 ! -o lo -j DROP
"""

network_disruption_ipv6 = network_disruption.replace('iptables', 'ip6tables')

undisrupt_network = """
 sudo iptables-restore ~/working.iptables.rules
"""

undisrupt_network_ipv6 = undisrupt_network.replace('iptables', 'ip6tables')

# TODO(eolivare): run ovn_db_pcs_resource_restart using
# run_pcs_resource_operation
# Now it is not possible because it is executed with
# ssh_client.connect().exec_command and run_pcs_resource_operation only
# supports sh.execute
ovn_db_pcs_resource_restart = (f"sudo pcs resource restart "
                               f"{pacemaker.OVN_DBS_RESOURCE}")
kill_rabbit = "sudo pkill -9 beam.smp"
remove_grastate = "sudo rm -rf /var/lib/mysql/grastate.dat"
check_bootstrap = """ps -eo lstart,cmd | grep -v grep|
grep wsrep-cluster-address=gcomm://"""
galera_sst_request = """sudo grep 'wsrep_sst_rsync.*'
/var/log/containers/mysql/mysqld.log"""
kill_mysqld = "sudo pkill -9 mysqld"
kill_mariadbd = "sudo pkill -9 mariadbd"


class PcsDisableException(tobiko.TobikoException):
    message = "pcs disable didn't shut down the resource"


class PcsEnableException(tobiko.TobikoException):
    message = "pcs enable didn't start the resource"


class GaleraBoostrapException(tobiko.TobikoException):
    message = "Bootstrap has not been activated"


class TimestampException(tobiko.TobikoException):
    message = "Timestamp mismatch: sst was requested before grastate removal"


def network_disrupt_node(node_name, disrupt_method=network_disruption):
    disrupt_node(node_name, disrupt_method=disrupt_method)


def network_undisrupt_node(node_name, disrupt_method=undisrupt_network):
    disrupt_node(node_name, disrupt_method=disrupt_method)


def disrupt_node(node_name, disrupt_method=network_disruption):
    # reboot all controllers and wait for ssh Up on them
    # hard reset is simultaneous while soft is sequential
    # method : method of disruption to use : network_disruption |
    # container_restart

    start_time = tobiko.time()
    # using ssh_client.connect we use a fire and forget reboot method
    node = tripleo_topology.get_node(node_name)
    node.ssh_client.connect().exec_command(disrupt_method)
    LOG.info('disrupt exec: {} on server: {}'.format(disrupt_method,
                                                     node.name))

    if isinstance(disrupt_method, sh.RebootHostMethod) \
            or is_network_disruption(disrupt_method):
        check_overcloud_node_uptime(node.ssh_client, start_time)
    else:
        check_overcloud_node_responsive(node)


def check_overcloud_node_uptime(ssh_client, start_time):
    for attempt in tobiko.retry(timeout=600., interval=10.):
        try:
            uptime = sh.get_uptime(ssh_client=ssh_client, timeout=15.)
        except sh.UptimeError:
            LOG.exception('uptime command failed')
            uptime = None

        if uptime and uptime < (tobiko.time() - start_time):
            LOG.debug('Reboot has been completed')
            break
        else:
            attempt.check_limits()


def reboot_node(node_name, wait=True, reboot_method=sh.hard_reset_method):

    # reboot a node and wait for ssh Up on them
    # hard reset is simultaneous while soft is sequential
    # method : method of disruption to use : reset | network_disruption

    # using ssh_client.connect we use a fire and forget reboot method
    node = tripleo_topology.get_node(node_name)
    sh.reboot_host(ssh_client=node.ssh_client, wait=wait, method=reboot_method)
    LOG.info('disrupt exec: {} on server: {}'.format(reboot_method,
                                                     node.name))


def check_overcloud_node_responsive(node):
    node_checked = sh.execute("hostname",
                              ssh_client=node.ssh_client,
                              expect_exit_status=None).stdout
    LOG.info('{} is up '.format(node_checked))

    tobiko.cleanup_fixture(node.ssh_client)


def network_disrupt_all_controller_nodes(disrupt_method=network_disruption,
                                         exclude_list=None):
    disrupt_all_controller_nodes(disrupt_method=disrupt_method,
                                 exclude_list=exclude_list)


def reset_all_controller_nodes(disrupt_method=sh.hard_reset_method,
                               exclude_list=None):
    disrupt_all_controller_nodes(disrupt_method=disrupt_method,
                                 exclude_list=exclude_list)


def reset_all_controller_nodes_sequentially(
        disrupt_method=sh.hard_reset_method,
        sequentially=True, exclude_list=None):
    disrupt_all_controller_nodes(disrupt_method=disrupt_method,
                                 sequentially=sequentially,
                                 exclude_list=exclude_list)


def disrupt_all_controller_nodes(disrupt_method=sh.hard_reset_method,
                                 sequentially=False, exclude_list=None):
    # TODO(eolivare): join disrupt_all_controller_nodes and
    # reboot_all_controller_nodes methods because they are very similar

    # reboot all controllers and wait for ssh Up on them
    # method : method of disruptino to use : reset | network_disruption
    # hard reset is simultaneous while soft is sequential
    # exclude_list = list of nodes to NOT reset

    controlplane_groups = ['controller', 'messaging', 'database', 'networker']
    actual_controlplane_groups = tripleo_topology.actual_node_groups(
        controlplane_groups)
    nodes = topology.list_openstack_nodes(group=actual_controlplane_groups)

    # remove excluded nodes from reset list
    if exclude_list:
        nodes = [node for node in nodes if node.name not in exclude_list]

    start_time = {}
    for controller in nodes:
        start_time[controller.name] = tobiko.time()
        if isinstance(disrupt_method, sh.RebootHostMethod):
            reboot_node(controller.name, wait=sequentially,
                        reboot_method=disrupt_method)
        else:
            # using ssh_client.connect we use a fire and forget reboot method
            controller.ssh_client.connect().exec_command(disrupt_method)
            LOG.info('disrupt exec: {} on server: {}'.format(disrupt_method,
                                                             controller.name))
            tobiko.cleanup_fixture(controller.ssh_client)
            if sequentially and is_network_disruption(disrupt_method):
                check_overcloud_node_uptime(
                    controller.ssh_client, start_time[controller.name])
        if sequentially:
            check_overcloud_node_responsive(controller)

    if not sequentially:
        for controller in nodes:
            if isinstance(disrupt_method, sh.RebootHostMethod) \
                    or is_network_disruption(disrupt_method):
                check_overcloud_node_uptime(
                    controller.ssh_client, start_time[controller.name])
            else:
                check_overcloud_node_responsive(controller)


def reboot_all_controller_nodes(reboot_method=sh.hard_reset_method,
                                sequentially=False, exclude_list=None):
    # reboot all controllers and wait for ssh Up on them
    # method : method of disruptino to use : hard or soft reset
    # hard reset is simultaneous while soft is sequential
    # exclude_list = list of nodes to NOT reset

    controlplane_groups = ['controller', 'messaging', 'database', 'networker']
    actual_controlplane_groups = tripleo_topology.actual_node_groups(
        controlplane_groups)
    nodes = topology.list_openstack_nodes(group=actual_controlplane_groups)

    # remove excluded nodes from reset list
    if exclude_list:
        nodes = [node for node in nodes if node.name not in exclude_list]

    start_time = {}
    for controller in nodes:
        start_time[controller.name] = tobiko.time()
        sh.reboot_host(ssh_client=controller.ssh_client, wait=sequentially,
                       method=reboot_method)
        LOG.info('reboot exec: {} on server: {}'.format(reboot_method,
                                                        controller.name))
        tobiko.cleanup_fixture(controller.ssh_client)
    if not sequentially:
        for controller in nodes:
            check_overcloud_node_uptime(
                controller.ssh_client, start_time[controller.name])


def is_ipv6addr_main_vip():
    """ Return True if main OC vip is ivp6 address. Otherwise return False."""
    main_vip = get_main_vip()
    return not netaddr.valid_ipv4(main_vip) and netaddr.valid_ipv6(main_vip)


def get_main_vip():
    """return the ip of the overcloud main vip.
    Retreive an ip address (ipv4/ipv6) from the auth_url."""
    auth_url = keystone.default_keystone_credentials().auth_url
    auth_url_parsed = urllib.parse.urlsplit(auth_url)
    main_vip = auth_url_parsed.hostname

    if not (netaddr.valid_ipv4(main_vip) or netaddr.valid_ipv6(main_vip)):
        try:
            # socket.gethostbyname translates hostname to IPv4 - it fails when
            # no IPv4 address is available
            main_vip = socket.gethostbyname(main_vip)
        except socket.gaierror:
            # the following method obtains an IPv6 from a hostname
            main_vip = socket.getaddrinfo(
                main_vip, None, socket.AF_INET6)[0][4][0]

    return main_vip


def get_main_vip_controller(main_vip):
    """return the controller hostname ,
    which is holding the main_vip pacemaker resource"""
    # when the main_vip is ipv6, the pacemaker command output replaces : by .
    # we need to adapt the value accordingly
    main_vim_controller = pacemaker.get_overcloud_nodes_running_pcs_resource(
        resource=f"ip-{main_vip.replace(':', '.')}")[0]
    return main_vim_controller


def delete_evacuable_tagged_image():
    # delete evacuable tagged image because it prevents
    # non tagged evacuations if exists
    for img in glance.list_images():
        if 'evacuable' in img['tags']:
            glance.delete_image(img.id)


def disrupt_controller_main_vip(disrupt_method=sh.hard_reset_method,
                                inverse=False):

    # reset the controller holding the main vip (os_auth_url)
    # ip resource (managed via pacemaker)
    # find main vip by getting it from
    main_vip = get_main_vip()

    # find the node holding that resource via :

    main_vip_controller = get_main_vip_controller(main_vip)

    if isinstance(disrupt_method, sh.RebootHostMethod):
        if inverse:
            reboot_all_controller_nodes(reboot_method=disrupt_method,
                                        exclude_list=[main_vip_controller])
        else:
            reboot_node(main_vip_controller, reboot_method=disrupt_method)
    else:
        if inverse:
            # inverse the nodes reset selection
            disrupt_all_controller_nodes(disrupt_method=disrupt_method,
                                         exclude_list=[main_vip_controller])
        else:
            # get that node's ssh_client and reset it
            disrupt_node(main_vip_controller, disrupt_method=disrupt_method)


def disrupt_controller_galera_main_vip(disrupt_method=sh.soft_reset_method):
    # This case reboots controller while VM creation is in progress
    # Please refer to RHBZ#2124877 for more info
    # Find the Galera VIP (port name : internal_api_virtual_ip)
    session = tripleo.undercloud_keystone_session()
    uc_neutron_client = neutron.get_neutron_client(session=session)
    try:
        new_port = neutron.find_port(client=uc_neutron_client, unique=False,
                                     name='internal_api_virtual_ip')
    except tobiko.ObjectNotFound as no_internal_api:
        raise tobiko.SkipException(
            'This OSP environment doesnt have an internal_api \
              network, so this test cannot be executed') from no_internal_api

    galera_vip_address = new_port['fixed_ips'][0]['ip_address']
    LOG.info("The Galera VIP address is: %r", galera_vip_address)
    # Find the controller hosting VIP resource
    galera_vip_resource = "ip-"+galera_vip_address
    galera_vip_controller = pacemaker.get_overcloud_nodes_running_pcs_resource(
                               resource=galera_vip_resource)[0]

    ports_before_stack_creation = neutron.list_ports(
        device_owner="compute:nova")
    multi_ip_test_fixture = tobiko.get_fixture(
        stacks.MultiIPCirrosServerStackFixture)
    tobiko.use_fixture(multi_ip_test_fixture)
    time.sleep(10)  # wait until some of the VMs have been created

    # Reboot that controller
    reboot_node(galera_vip_controller, wait=True,
                reboot_method=disrupt_method)

    return multi_ip_test_fixture, ports_before_stack_creation


def get_vms_detailed_info(multi_ip_test_fixture):
    for attempt in tobiko.retry(timeout=300, interval=10):
        # dynamically obtain the status of the VMs
        vms_detailed_info = multi_ip_test_fixture.vms_detailed_info

        vm_status_list = [
            vm.get('status') for vm in vms_detailed_info if vm is not None]
        if 'BUILD' not in vm_status_list:
            LOG.debug("All VMs reached a final status")
            break
        if attempt.is_last:
            LOG.warn("Still some VMs in status BUILD - the test continues...")
            break

    return vms_detailed_info


def check_no_duplicate_ips(vms_detailed_info, ports_before_stack_creation):
    test_case = tobiko.get_test_case()
    # check VM IP addresses are different
    ip4_list = []
    ip6_list = []
    for vm in vms_detailed_info:
        addresses = vm.get('addresses', {}) if vm is not None else {}
        # try to obtain the port associated to a VM from neutron if the VM
        # exists but vms_detailed_info does not show the port
        if not addresses and vm is not None:
            ports = neutron.list_ports(device_id=vm['id'],
                                       device_owner="compute:nova")
            test_case.assertLess(len(ports), 2)
            for port in ports:
                addresses[port['network_id']] = port['fixed_ips']

        for addresses_per_network in addresses.values():
            test_case.assertEqual(len(addresses_per_network), 2)
            for subnet_addr in addresses_per_network:
                # the subnet_addr dict is different depending on how it was
                # obtained: from vms_detailed_info or from neutron.list_ports
                subnet_ip = (subnet_addr.get('addr') or
                             subnet_addr.get('ip_address'))
                if netaddr.valid_ipv4(subnet_ip):
                    ip4_list.append(subnet_ip)
                elif netaddr.valid_ipv6(subnet_ip):
                    ip6_list.append(subnet_ip)

    ip4_set = set(ip4_list)  # this removes duplicate values
    LOG.debug("list of IPv4s from the MultiIPVM group: %r", ip4_list)
    test_case.assertEqual(len(ip4_list), len(ip4_set))

    ip6_set = set(ip6_list)  # this removes duplicate values
    LOG.debug("list of IPv6s from the MultiIPVM group: %r", ip6_list)
    test_case.assertEqual(len(ip6_list), len(ip6_set))

    LOG.debug("list of IPv4 and list of IPv6 addresses "
              "should have the same length")
    test_case.assertEqual(len(ip6_list), len(ip4_list))

    ports_after_reboot = neutron.list_ports(device_owner="compute:nova")
    ports_after_reboot_full = neutron.list_ports()
    LOG.debug(
        "list of compute:nova ports obtained at the begining of this test: %r",
        ports_before_stack_creation)
    LOG.debug(
        "list of compute:nova ports obtained at the end of this test: %r",
        ports_after_reboot)
    LOG.debug("full list of ports obtained at the end of this test: %r",
              ports_after_reboot_full)
    test_case.assertEqual(len(ip6_list), len(ports_after_reboot) - len(
        ports_before_stack_creation))


def reboot_controller_galera_main_vip():
    return disrupt_controller_galera_main_vip(
        disrupt_method=sh.soft_reset_method)


def reset_controller_main_vip():
    disrupt_controller_main_vip(disrupt_method=sh.hard_reset_method)


def reset_controllers_non_main_vip():
    disrupt_controller_main_vip(disrupt_method=sh.hard_reset_method,
                                inverse=True)


def crash_controller_main_vip():
    disrupt_controller_main_vip(disrupt_method=sh.crash_method)


def crash_controllers_non_main_vip():
    disrupt_controller_main_vip(disrupt_method=sh.crash_method,
                                inverse=True)


def network_disrupt_controller_main_vip():
    disrupt_controller_main_vip(disrupt_method=get_network_disruption(
                                is_ipv6addr_main_vip()))
    LOG.info('waiting 60s to avoid race conditions...')
    time.sleep(60.0)


def network_undisrupt_controller_main_vip():
    disrupt_controller_main_vip(disrupt_method=get_undisrupt_network(
                                is_ipv6addr_main_vip()))


def network_disrupt_controllers_non_main_vip():
    disrupt_controller_main_vip(disrupt_method=get_network_disruption(
                                is_ipv6addr_main_vip()),
                                inverse=True)


def network_undisrupt_controllers_non_main_vip():
    disrupt_controller_main_vip(disrupt_method=get_undisrupt_network(
                                is_ipv6addr_main_vip()),
                                inverse=True)


def is_network_disruption(disrupt_method):
    if re.search("ip6?tables", disrupt_method) \
                 and not re.search("restore", disrupt_method):
        return True
    else:
        return False


def get_network_disruption(isIpv6=False):
    """
    Param: isIpv6: boolean
    Return network_disruption(for ipv4) or network_disruption_ipv6 (which are
    iptables or ip6tables rules, correspondingly), according to the
    isIpv6 flag.
    """
    if isIpv6:
        return network_disruption_ipv6
    else:
        return network_disruption


def get_undisrupt_network(isIpv6=False):
    """
    Return undisrupt_network(for ipv4) or undisrupt_network_ipv6 method,
    according to the isIpv6 flag.
    """
    if isIpv6:
        return undisrupt_network_ipv6
    else:
        return undisrupt_network


def reset_all_compute_nodes(hard_reset=False, sequentially=False):

    # reboot all computes and wait for ssh Up on them
    if hard_reset:
        reset_method = sh.hard_reset_method
    else:
        reset_method = sh.soft_reset_method

    nodes = topology.list_openstack_nodes(group='compute')
    compute_reboot_operation_list = []
    for compute in nodes:
        # using ssh_client.connect we use a fire and forget reboot method
        # if sequentially, then wait and check uptime has changed
        # else, do not wait (uptime will be checked later)
        reboot_operation = sh.reboot_host(ssh_client=compute.ssh_client,
                                          wait=sequentially,
                                          method=reset_method)
        compute_reboot_operation_list.append(reboot_operation)
        LOG.info('reboot exec:  {} on server: {}'.format(reset_method,
                                                         compute.name))

    if not sequentially:
        for reboot_operation in compute_reboot_operation_list:
            # checking uptime on each compute - it should have been updated
            # after the reboot is done
            reboot_operation.wait_for_operation()
            LOG.info(f'{reboot_operation.hostname} is up')


def reset_ovndb_pcs_master_resource():
    """restart ovndb pacemaker resource
    this method only restart the resource running on the controller with is
    acting as Master"""
    node = pacemaker.get_overcloud_nodes_running_pcs_resource(
        resource_type='(ocf::ovn:ovndb-servers):', resource_state='Master')[0]
    ovn_db_pcs_master_resource_restart = (ovn_db_pcs_resource_restart + ' ' +
                                          node)
    disrupt_node(node, disrupt_method=ovn_db_pcs_master_resource_restart)


def reset_ovndb_pcs_resource():
    """restart ovndb pacemaker resource
    this method restart the whole resource, i.e. on all the controller nodes"""
    node = pacemaker.get_overcloud_nodes_running_pcs_resource(
        resource_type='(ocf::ovn:ovndb-servers):', resource_state='Master')[0]
    disrupt_node(node, disrupt_method=ovn_db_pcs_resource_restart)


def reset_ovndb_master_container():
    """get and restart the ovndb master container
    use of partial name :  resource: ovn-dbs-bundle-0 =>
    container: ovn-dbs-bundle-podman-0 or ovn-dbs-bundle-docker-0"""
    node = pacemaker.get_overcloud_nodes_running_pcs_resource(
        resource_type='(ocf::ovn:ovndb-servers):', resource_state='Master')[0]
    resource = pacemaker.get_overcloud_resource(
        resource_type='(ocf::ovn:ovndb-servers):', resource_state='Master')
    resource = resource[0].rsplit('-', 1)[0]
    containers.action_on_container('restart',
                                   partial_container_name=resource,
                                   container_host=node)


def restart_service_on_all_nodes(service):
    """restart the ovn bgp agent or the frr service from all the nodes where it
    is running and check the cloud is healthy after they are started again"""
    node_names = tripleo.get_overcloud_nodes_running_service(service)
    nodes = topology.list_openstack_nodes(hostnames=node_names)
    for node in nodes:
        sh.stop_systemd_units(service, ssh_client=node.ssh_client)
    for node in nodes:
        sh.start_systemd_units(service, ssh_client=node.ssh_client)


def kill_rabbitmq_service():
    """kill a rabbit process on a random controller,
    check in pacemaker it is down"""
    if tripleo_topology.is_composable_roles_env():
        nodes = topology.list_openstack_nodes(group='messaging')
    else:
        nodes = topology.list_openstack_nodes(group='controller')
    node = random.choice(nodes)
    sh.execute(kill_rabbit, ssh_client=node.ssh_client)
    LOG.info('kill rabbit: {} on server: {}'.format(kill_rabbit,
             node.name))
    retry = tobiko.retry(timeout=30, interval=5)
    for _ in retry:
        if not(pacemaker.PacemakerResourcesStatus().
               rabbitmq_resource_healthy()):
            return


def kill_all_galera_services():
    """kill all galera processes,
    check in pacemaker it is down"""
    if tripleo_topology.is_composable_roles_env():
        nodes = topology.list_openstack_nodes(group='database')
    else:
        nodes = topology.list_openstack_nodes(group='controller')
    for node in nodes:
        if topology.verify_osp_version('17.0', lower=True):
            sh.execute(kill_mysqld, ssh_client=node.ssh_client)
            LOG.info('kill galera: {} on server: {}'.format(kill_mysqld,
                                                            node.name))
        else:
            sh.execute(kill_mariadbd, ssh_client=node.ssh_client)
            LOG.info('kill galera: {} on server: {}'.format(kill_mariadbd,
                                                            node.name))
    retry = tobiko.retry(timeout=30, interval=5)
    for _ in retry:
        if not(pacemaker.PacemakerResourcesStatus().
               galera_resource_healthy()):
            return


def remove_all_grastate_galera():
    """shut down galera properly,
    remove all grastate"""
    if tripleo_topology.is_composable_roles_env():
        nodes = topology.list_openstack_nodes(group='database')
    else:
        nodes = topology.list_openstack_nodes(group='controller')
    LOG.info('shut down {} on all servers: {}'.format(
        pacemaker.GALERA_RESOURCE, nodes))
    if f"resource '{pacemaker.GALERA_RESOURCE}' is not running on any node" \
            not in pacemaker.run_pcs_resource_operation(
                pacemaker.GALERA_RESOURCE,
                pacemaker.DISABLE,
                nodes[0].ssh_client):
        raise PcsDisableException()
    for node in nodes:
        sh.execute(remove_grastate, ssh_client=node.ssh_client)

    LOG.info('enable back {} on all servers: {}'.format(
        pacemaker.GALERA_RESOURCE, nodes))
    if topology.verify_osp_version('17.0', lower=True):
        promoted = "master"
    else:
        promoted = "promoted"
    if f"resource '{pacemaker.GALERA_RESOURCE}' is {promoted} on node" not in \
            pacemaker.run_pcs_resource_operation(pacemaker.GALERA_RESOURCE,
                                                 pacemaker.ENABLE,
                                                 nodes[0].ssh_client,
                                                 operation_wait=90):
        raise PcsEnableException()


def remove_one_grastate_galera():
    """shut down galera properly,
    delete /var/lib/mysql/grastate.dat in a random node,
    check that bootstrap is done from a node with grastate"""
    if tripleo_topology.is_composable_roles_env():
        nodes = topology.list_openstack_nodes(group='database')
    else:
        nodes = topology.list_openstack_nodes(group='controller')
    node = random.choice(nodes)

    pcs_haproxy = pacemaker.HAPROXY_RESOURCE
    pcs_galera = pacemaker.GALERA_RESOURCE

    if not CONF.tobiko.rhosp.has_external_load_balancer:
        LOG.info(f'disable {pcs_haproxy}')
        if f"resource '{pcs_haproxy}' is not running on any node" not in \
                pacemaker.run_pcs_resource_operation(
                    pcs_haproxy,
                    pacemaker.DISABLE,
                    node.ssh_client,
                    operation_wait=30):
            raise PcsDisableException()
    else:
        LOG.debug(f'With Ext LB setups, {pcs_haproxy} is not deployed')

    LOG.info('shut down {} on all servers: {}'.format(pcs_galera, nodes))
    if f"resource '{pcs_galera}' is not running on any node" not in \
            pacemaker.run_pcs_resource_operation(
                pcs_galera,
                pacemaker.DISABLE,
                node.ssh_client):
        raise PcsDisableException()
    LOG.info('remove grastate: {} on server: {}'.format(remove_grastate,
                                                        node.name))
    sh.execute(remove_grastate, ssh_client=node.ssh_client)

    LOG.info('enable back {} on all servers: {}'.format(pcs_galera, nodes))
    if topology.verify_osp_version('17.0', lower=True):
        promoted = "master"
    else:
        promoted = "promoted"
    if f"resource '{pcs_galera}' is {promoted} on node" not in \
            pacemaker.run_pcs_resource_operation(
                pcs_galera, pacemaker.ENABLE, node.ssh_client,
                operation_wait=90):
        raise PcsEnableException()

    if not CONF.tobiko.rhosp.has_external_load_balancer:
        LOG.info(f'enable {pcs_haproxy}')
        if f"resource '{pcs_haproxy}' is running on node" not in \
                pacemaker.run_pcs_resource_operation(pcs_haproxy,
                                                     pacemaker.ENABLE,
                                                     node.ssh_client):
            raise PcsEnableException()
    else:
        LOG.debug(f'With Ext LB setups, {pcs_haproxy} is not deployed')

    # gcomm:// without args means that bootstrap is done from this node
    bootstrap = sh.execute(check_bootstrap, ssh_client=node.ssh_client).stdout
    if re.search('wsrep-cluster-address=gcomm://', bootstrap) is None:
        raise GaleraBoostrapException()
    lastDate = re.findall(r"\w{,3}\s*\w{,3}\s*\d{,2}\s*\d{,2}:\d{,2}:\d{,2}\s*"
                          r"\d{4}", bootstrap)[-1]
    return node, lastDate


def request_galera_sst():
    """remove_one_grastate_galera,
    check that sst is requested by a node with grastate"""
    node, date = remove_one_grastate_galera()
    bootstrapDate = datetime.strptime(date, '%a %b %d %H:%M:%S %Y')
    retry = tobiko.retry(timeout=30, interval=5)
    for _ in retry:
        sst_req = sh.execute(galera_sst_request,
                             ssh_client=node.ssh_client).stdout
        if sst_req:
            break
    sstDate = datetime.strptime(re.findall
                                (r"\d{4}-\d{,2}-\d{,2}\s*\d{,2}:\d{,2}:\d{,2}",
                                 sst_req)[-1], '%Y-%m-%d %H:%M:%S')
    if bootstrapDate > sstDate:
        raise TimestampException


def evac_failover_compute(compute_host, failover_type=sh.hard_reset_method):
    """disrupt a compute, to trigger it's instance-HA evacuation
    failover_type=hard_reset_method etc.."""
    if failover_type in (sh.hard_reset_method, sh.soft_reset_method):
        reboot_node(compute_host, reboot_method=failover_type)
    else:
        disrupt_node(compute_host, disrupt_method=failover_type)


def check_iha_evacuation(failover_type=None, vm_type=None):
    """check vms on compute host,disrupt compute host,
    check all vms evacuated and pingable"""
    for iteration in range(2):
        LOG.info(f'Begin IHA tests iteration {iteration}')
        LOG.info('create 2 vms')
        tests.test_servers_creation(number_of_servers=2)
        compute_host = nova.get_random_compute_with_vms_name()
        vms_starting_state_df = nova.get_compute_vms_df(compute_host)
        if vm_type == 'shutoff':
            nova.stop_all_instances()
        if vm_type == 'evac_image_vm':
            evac_vm_stack = tests.test_evacuable_server_creation()
            evac_vm_id = nova.get_stack_server_id(evac_vm_stack)
            org_nova_evac_df = nova.vm_df(evac_vm_id, nova.get_vms_table())
        if not vm_type == 'shutoff':
            nova.check_df_vms_ping(vms_starting_state_df)
        LOG.info(f'perform a failover on {compute_host}')
        evac_failover_compute(compute_host, failover_type=failover_type)
        test_cloud_recovery.overcloud_health_checks(passive_checks_only=True)
        if vm_type == 'evac_image_vm':
            nova.check_vm_evacuations(vms_df_old=org_nova_evac_df,
                                      compute_host=compute_host,
                                      timeout=600,
                                      check_no_evacuation=True)
            # delete evacuable tagged image because it prevents
            # non tagged evacuations if exists
            delete_evacuable_tagged_image()
            new_nova_evac_df = nova.vm_df(evac_vm_id, nova.get_vms_table())
            nova.check_vm_evacuations(org_nova_evac_df, new_nova_evac_df)
        else:
            nova.check_vm_evacuations(vms_df_old=vms_starting_state_df,
                                      compute_host=compute_host,
                                      timeout=600)
        LOG.info('check evac is Done')
        if not vm_type == 'shutoff':
            nova.check_df_vms_ping(vms_starting_state_df)


def check_iha_evacuation_evac_image_vm():
    check_iha_evacuation(failover_type=sh.hard_reset_method,
                         vm_type='evac_image_vm')


def check_iha_evacuation_hard_reset():
    check_iha_evacuation(failover_type=sh.hard_reset_method)


def check_iha_evacuation_network_disruption():
    check_iha_evacuation(failover_type=get_network_disruption(
                         is_ipv6addr_main_vip()))


def check_iha_evacuation_hard_reset_shutoff_instance():
    check_iha_evacuation(failover_type=sh.hard_reset_method, vm_type='shutoff')


def test_controllers_shutdown():
    test_case = tobiko.get_test_case()

    all_nodes = topology.list_openstack_nodes(group='controller')
    if len(all_nodes) < 3:
        tobiko.skip_test('It requires at least three controller nodes')

    all_node_names = [node.name for node in all_nodes]
    LOG.info("Ensure all controller nodes are running: "
             f"{all_node_names}")
    for node in all_nodes:
        node.power_on_node()
    topology.assert_reachable_nodes(all_nodes)

    LOG.debug('Check VM is running while all controllers nodes are on')
    nova_server = tobiko.setup_fixture(stacks.CirrosServerStackFixture)
    nova_server_ip = nova_server.ip_address
    ping.assert_reachable_hosts([nova_server_ip])

    quorum_level = math.ceil(0.5 * len(all_nodes))
    assert quorum_level >= len(all_nodes) - quorum_level
    nodes = random.sample(all_nodes, quorum_level)
    node_names = [node.name for node in nodes]
    LOG.info(f"Power off {quorum_level} random controller nodes: "
             f"{node_names}")
    for node in nodes:
        node.power_off_node()
        test_case.addCleanup(node.power_on_node)
    topology.assert_unreachable_nodes(nodes, retry_count=1)
    topology.assert_reachable_nodes(node
                                    for node in all_nodes
                                    if node not in nodes)

    LOG.debug('Check whenever VM is still running while some "'
              '"controllers nodes are off')
    reachable, unreachable = ping.ping_hosts([nova_server_ip],
                                             count=1)
    if reachable:
        LOG.debug(f"VM ips are reachable: {reachable}")
    if unreachable:
        LOG.debug(f"VM is are unreachable: {unreachable}")
    # TODO what do we expect here: VM reachable or unreachable?

    random.shuffle(nodes)
    LOG.info(f"Power on controller nodes: {node_names}")
    for node in nodes:
        node.power_on_node()

    LOG.debug("Check all controller nodes are running again: "
              f"{all_node_names}")
    topology.assert_reachable_nodes(all_nodes, retry_timeout=600.)

    LOG.debug('Check VM is running while all controllers nodes are on')
    ping.assert_reachable_hosts([nova_server_ip])



from __future__ import absolute_import

import functools
import os
import re
import time
import typing

from oslo_log import log
import pandas

import tobiko
from tobiko import config
from tobiko import podman
from tobiko.openstack import neutron
from tobiko.openstack import topology
from tobiko import rhosp as rhosp_topology
from tobiko.rhosp import containers as rhosp_containers
from tobiko.shell import sh
from tobiko.tripleo import overcloud


CONF = config.CONF
LOG = log.getLogger(__name__)


class ContainerRuntimeFixture(tobiko.SharedFixture):

    runtime: typing.Optional[rhosp_containers.ContainerRuntime] = None

    def setup_fixture(self):
        if overcloud.has_overcloud():
            self.runtime = self.get_runtime()

    def cleanup_fixture(self):
        self.runtime = None

    @staticmethod
    def get_runtime() -> typing.Optional[rhosp_containers.ContainerRuntime]:
        """check what container runtime is running
        and return a handle to it"""
        # TODO THIS LOCKS SSH CLIENT TO CONTROLLER
        for node in topology.list_openstack_nodes(group='controller'):
            try:
                result = sh.execute('podman --version || docker --version',
                                    ssh_client=node.ssh_client)
            except sh.ShellCommandFailed:
                continue
            for runtime in rhosp_containers.CONTAINER_RUNTIMES:
                for version in [result.stdout, result.stderr]:
                    if runtime.match_version(version):
                        return runtime
        raise RuntimeError(
            "Unable to find any container runtime in any overcloud "
            "controller node")


def get_container_runtime() -> rhosp_containers.ContainerRuntime:
    runtime = tobiko.setup_fixture(ContainerRuntimeFixture).runtime
    return runtime


def get_container_runtime_name() -> str:
    return get_container_runtime().runtime_name


def is_docker() -> bool:
    return get_container_runtime().runtime_name == 'docker'


def is_podman() -> bool:
    return get_container_runtime().runtime_name == 'podman'


def has_container_runtime() -> bool:
    return get_container_runtime() is not None


def skip_unless_has_container_runtime():
    return tobiko.skip_unless('Container runtime not found',
                              has_container_runtime)


@functools.lru_cache()
def list_node_containers(ssh_client):
    """returns a list of containers and their run state"""
    return get_container_runtime().list_containers(ssh_client=ssh_client)


def get_container_client(ssh_client=None):
    """returns a list of containers and their run state"""
    return get_container_runtime().get_client(ssh_client=ssh_client)


def list_containers_df(group=None):
    actual_containers_list = list_containers(group)
    return pandas.DataFrame(
        get_container_states_list(actual_containers_list),
        columns=['container_host', 'container_name', 'container_state'])


def list_containers(group=None):
    """get list of containers in running state
    from specified node group
    returns : a list of overcloud_node's running containers"""

    # moved here from topology
    # reason : Workaround for :
    # AttributeError: module 'tobiko.openstack.topology' has no
    # attribute 'container_runtime'

    if group is None:
        group = 'overcloud'
    containers_list = tobiko.Selection()
    openstack_nodes = topology.list_openstack_nodes(group=group)

    for node in openstack_nodes:
        LOG.debug(f"List containers for node {node.name}")
        node_containers_list = list_node_containers(ssh_client=node.ssh_client)
        containers_list.extend(node_containers_list)
    return containers_list


def save_containers_state_to_file(expected_containers_list,):
    expected_containers_list_df = pandas.DataFrame(
        get_container_states_list(expected_containers_list),
        columns=['container_host', 'container_name', 'container_state'])
    expected_containers_list_df.to_csv(
        rhosp_containers.expected_containers_file)
    return rhosp_containers.expected_containers_file


def assert_containers_running(group, expected_containers, full_name=True,
                              bool_check=False, nodenames=None):

    """assert that all containers specified in the list are running
    on the specified openstack group(controller or compute etc..)
    if bool_check is True then return only True or false without failing"""

    if is_docker():
        LOG.info('not checking common containers since we are on docker')
        return

    failures = []

    openstack_nodes = topology.list_openstack_nodes(group=group,
                                                    hostnames=nodenames)
    for node in openstack_nodes:
        node_containers = list_node_containers(ssh_client=node.ssh_client)
        containers_list_df = pandas.DataFrame(
            get_container_states_list(node_containers),
            columns=['container_host', 'container_name', 'container_state'])
        # check that the containers are present
        LOG.info('node: {} containers list : {}'.format(
            node.name, containers_list_df.to_string(index=False)))
        for container in expected_containers:
            # get container attrs dataframe
            if full_name:
                container_attrs = containers_list_df.query(
                    'container_name == "{}"'.format(container))
            else:
                container_attrs = containers_list_df[
                    containers_list_df['container_name'].
                    str.contains(container)]
            # check if the container exists
            LOG.info('checking container: {}'.format(container))
            if container_attrs.empty:
                failures.append(
                    'expected container {} not found on node {} ! : \n\n'.
                    format(container, node.name))
            # if container exists, check it is running
            else:
                # only one running container is expected
                container_running_attrs = container_attrs.query(
                    'container_state=="running"')
                if container_running_attrs.empty:
                    failures.append(
                        'expected container {} is not running on node {} , '
                        'its state is {}! : \n\n'.format(
                            container, node.name,
                            container_attrs.container_state.values.item()))
                elif len(container_running_attrs) > 1:
                    failures.append(
                        'only one running container {} was expected on '
                        'node {}, but got {}! : \n\n'.format(
                            container, node.name,
                            len(container_running_attrs)))

    if not bool_check and failures:
        tobiko.fail(
            'container states mismatched:\n{}'.format('\n'.join(failures)),
            rhosp_containers.ContainerMismatchException)

    elif bool_check and failures:
        return False

    else:
        LOG.info('All specified containers are in running state! ')
        return True


def get_libvirt_container_name():
    if topology.verify_osp_version('17.0', lower=True):
        return 'nova_libvirt'
    else:
        return 'nova_virtqemud'


def assert_all_tripleo_containers_running():
    """check that all common tripleo containers are running
    param: group controller or compute , check containers
    sets in computes or controllers"""
    nova_libvirt = get_libvirt_container_name()

    common_controller_tripleo_containers = ['cinder_api', 'cinder_api_cron',
                                            'cinder_scheduler',
                                            'glance_api', 'heat_api',
                                            'heat_api_cfn',
                                            'heat_api_cron', 'heat_engine',
                                            'horizon', 'iscsid', 'keystone',
                                            'logrotate_crond', 'memcached',
                                            'neutron_api', 'nova_api',
                                            'nova_api_cron', 'nova_conductor',
                                            'nova_metadata', 'nova_scheduler',
                                            'nova_vnc_proxy',
                                            'swift_account_auditor',
                                            'swift_account_reaper',
                                            'swift_account_replicator',
                                            'swift_account_server',
                                            'swift_container_auditor',
                                            'swift_container_replicator',
                                            'swift_container_server',
                                            'swift_container_updater',
                                            'swift_object_auditor',
                                            'swift_object_expirer',
                                            'swift_object_replicator',
                                            'swift_object_server',
                                            'swift_object_updater',
                                            'swift_proxy', 'swift_rsync']

    shiftonstack_controller_tripleo_containers = ['cinder_api',
                                                  'cinder_api_cron',
                                                  'cinder_scheduler',
                                                  'glance_api', 'heat_api',
                                                  'heat_api_cfn',
                                                  'heat_api_cron',
                                                  'heat_engine',
                                                  'horizon', 'iscsid',
                                                  'keystone',
                                                  'logrotate_crond',
                                                  'memcached',
                                                  'neutron_api', 'nova_api',
                                                  'nova_api_cron',
                                                  'nova_conductor',
                                                  'nova_metadata',
                                                  'nova_scheduler',
                                                  'nova_vnc_proxy']

    common_compute_tripleo_containers = ['iscsid', 'logrotate_crond',
                                         'nova_compute', nova_libvirt,
                                         'nova_migration_target',
                                         'nova_virtlogd']

    if ceph_rgw_expected():
        tripleo_containers_to_check = \
            shiftonstack_controller_tripleo_containers
    else:
        tripleo_containers_to_check = \
            common_controller_tripleo_containers

    for group, group_containers in [('controller',
                                     tripleo_containers_to_check),
                                    ('compute',
                                     common_compute_tripleo_containers)]:
        assert_containers_running(group, group_containers)
    # TODO: need to address OSP-version specific containers here.
    # optional ovn containers checks
    assert_ovn_containers_running()


def assert_ceph_rgw_container_running():
    """check if ceph-rgw is running, usually on
    shifton stack deployments , if so we should not check for
    swift containers precence"""
    return assert_containers_running('controller', ['ceph-rgw'],
                                     full_name=False, bool_check=True)


def ceph_rgw_expected():
    """Returns True if 'ceph_rgw' parameter in the config file is set to 'true'
    or ceph rgw containers were detected running by Tobiko.
    """
    return CONF.tobiko.rhosp.ceph_rgw or assert_ceph_rgw_container_running()


def skip_if_ceph_rgw():
    return tobiko.skip_if('Ceph rgw deployed', ceph_rgw_expected)


def osp13_container_name_short_format(container_name_long_format):
    """This takes a long format container name :
    'rhosp13/openstack-neutron-l3-agent'
    and turns it into : neutron_l3_agent
    """
    return re.sub('-', '_', re.sub('rhosp13/openstack-', '',
                                   container_name_long_format))


def assert_ovn_containers_running():
    if not neutron.has_ovn():
        LOG.info("Networking OVN not configured")
        return

    # specific OVN verifications
    ovn_controller_containers = ['ovn_controller']
    if overcloud.is_ovn_using_raft():
        ovn_controller_containers += ['ovn_cluster_northd',
                                      'ovn_cluster_north_db_server',
                                      'ovn_cluster_south_db_server']
    else:
        container_runtime_name = get_container_runtime_name()
        ovn_controller_containers += [
            f'ovn-dbs-bundle-{container_runtime_name}-']
    ovn_compute_containers = ['ovn_metadata_agent',
                              'ovn_controller']
    group_containers_list = [('controller', ovn_controller_containers),
                             ('compute', ovn_compute_containers)]
    if 'networker' in topology.list_openstack_node_groups():
        ovn_networker_containers = ['ovn_controller']
        group_containers_list.append(('networker',
                                      ovn_networker_containers))
    for group, group_containers in group_containers_list:
        assert_containers_running(group, group_containers, full_name=False)
    LOG.info("Networking OVN containers verified in running state")


def run_container_config_validations():
    """check containers configuration in different scenarios
    """

    # TODO add here any generic configuration validation
    config_checkings = []

    if neutron.has_ovn():
        ovn_config_checkings = \
            [{'node_group': 'controller',
              'container_name': 'neutron_api',
              'config_file': '/etc/neutron/plugins/ml2/ml2_conf.ini',
              'param_validations': [{'section': 'ml2',
                                     'param': 'mechanism_drivers',
                                     'expected_value': 'ovn'},
                                    {'section': 'ml2',
                                     'param': 'type_drivers',
                                     'expected_value': 'geneve'},
                                    {'section': 'ovn',
                                     'param': 'ovn_metadata_enabled',
                                     'expected_value': 'True'}]}]
        config_checkings += ovn_config_checkings
    else:
        ovs_config_checkings = \
            [{'node_group': 'controller',
              'container_name': 'neutron_api',
              'config_file': '/etc/neutron/plugins/ml2/ml2_conf.ini',
              'param_validations': [{'section': 'ml2',
                                     'param': 'mechanism_drivers',
                                     'expected_value': 'openvswitch'}]}]
        config_checkings += ovs_config_checkings

    container_runtime_name = get_container_runtime_name()
    for config_check in config_checkings:
        for node in topology.list_openstack_nodes(
                group=config_check['node_group']):
            for param_check in config_check['param_validations']:
                obtained_param = sh.execute(
                    f"{container_runtime_name} exec -uroot "
                    f"{config_check['container_name']} crudini "
                    f"--get {config_check['config_file']} "
                    f"{param_check['section']} {param_check['param']}",
                    ssh_client=node.ssh_client, sudo=True).stdout.strip()
                if param_check['expected_value'] not in obtained_param:
                    tobiko.fail(f"Expected {param_check['param']} value: "
                                f"{param_check['expected_value']}\n"
                                f"Obtained {param_check['param']} value: "
                                f"{obtained_param}")
        LOG.info("Configuration verified:\n"
                 f"node group: {config_check['node_group']}\n"
                 f"container: {config_check['container_name']}\n"
                 f"config file: {config_check['config_file']}")


def comparable_container_keys(container, include_container_objects=False):
    """returns the tuple : 'container_host','container_name',
    'container_state, container object if specified'
     """
    # Differenciate between podman_ver3 with podman-py from earlier api
    if is_podman():
        if podman.Podman_Version_3():
            con_host_name_stat_obj_tuple = (rhosp_topology.ip_to_hostname(
                container.client.base_url.netloc.rsplit('_')[1]),
                                            container.attrs[
                'Names'][0], container.attrs['State'], container)

            con_host_name_stat_tuple = (rhosp_topology.ip_to_hostname(
                container.client.base_url.netloc.rsplit('_')[1]),
                                        container.attrs[
                'Names'][0], container.attrs['State'])
        else:

            con_host_name_stat_obj_tuple = (rhosp_topology.ip_to_hostname(
                container._client._context.hostname),  # pylint: disable=W0212
                container.data['names'], container.data['status'], container)

            con_host_name_stat_tuple = (rhosp_topology.ip_to_hostname(
                container._client._context.hostname),  # pylint: disable=W0212
                container.data['names'], container.data['status'])

        if include_container_objects:
            return con_host_name_stat_obj_tuple
        else:
            return con_host_name_stat_tuple

    elif is_docker():
        if include_container_objects:
            return (container.client.api.ssh_client.hostname,
                    osp13_container_name_short_format(container.attrs[
                                                          'Labels']['name']),
                    container.attrs['State'],
                    container)
        else:
            return (container.client.api.ssh_client.hostname,
                    osp13_container_name_short_format(container.attrs[
                                                          'Labels']['name']),
                    container.attrs['State'])


@functools.lru_cache()
def list_containers_objects_df():
    containers_list = list_containers()
    containers_objects_list_df = pandas.DataFrame(
        get_container_states_list(
            containers_list, include_container_objects=True),
        columns=['container_host', 'container_name',
                 'container_state', 'container_object'])
    return containers_objects_list_df


def get_overcloud_container(container_name=None, container_host=None,
                            partial_container_name=None):
    """gets an container object by name on specified host
    container"""
    con_obj_df = list_containers_objects_df()
    if partial_container_name and container_host:
        con_obj_df = con_obj_df[con_obj_df['container_name'].str.contains(
            partial_container_name)]
        contaniner_obj = con_obj_df.query(
            'container_host == "{container_host}"'.format(
                container_host=container_host))['container_object']
    elif container_host:
        contaniner_obj = con_obj_df.query(
            'container_name == "{container_name}"'
            ' and container_host == "{container_host}"'.
            format(container_host=container_host,
                   container_name=container_name)).container_object
    else:
        contaniner_obj = con_obj_df.query(
            'container_name == "{container_name}"'.
            format(container_name=container_name)).container_object
    if not contaniner_obj.empty:
        return contaniner_obj.values[0]
    else:
        tobiko.fail('container {} not found!'.format(container_name))


def action_on_container(action: str,
                        container_name=None,
                        container_host=None,
                        partial_container_name=None):
    """take a container snd preform an action on it
    actions are as defined in : podman/libs/containers.py:14/164"""

    LOG.debug(f"Executing '{action}' action on container "
              f"'{container_name}@{container_host}'...")
    container = get_overcloud_container(
        container_name=container_name,
        container_host=container_host,
        partial_container_name=partial_container_name)

    container_class: typing.Type = type(container)
    # we get the specified action as function from podman lib
    action_method: typing.Optional[typing.Callable] = getattr(
        container_class, action, None)
    if action_method is None:
        raise TypeError(f"Unsupported container action for class :"
                        f" {container_class}")
    if not callable(action_method):
        raise TypeError(
            f"Attribute '{container_class.__qualname__}.{action}' value "
            f" is not a method: {action_method!r}")
    LOG.debug(f"Calling '{action_method}' action on container "
              f"'{container}'")
    return action_method(container)


def get_container_states_list(containers_list,
                              include_container_objects=False):
    container_states_list = tobiko.Selection()
    container_states_list.extend([comparable_container_keys(
        container, include_container_objects=include_container_objects) for
                                  container in containers_list])
    return container_states_list


def assert_equal_containers_state(expected_containers_list=None,
                                  timeout=120, interval=2,
                                  recreate_expected=False):

    """compare all overcloud container states with using two lists:
    one is current , the other some past list
    first time this method runs it creates a file holding overcloud
    containers' states: ~/expected_containers_list_df.csv'
    second time it creates a current containers states list and
    compares them, they must be identical"""

    expected_containers_list_df = []
    # if we have a file or an explicit variable use that , otherwise  create
    # and return
    if recreate_expected or (
            not expected_containers_list and
            not os.path.exists(rhosp_containers.expected_containers_file)):
        save_containers_state_to_file(list_containers())
        return

    elif expected_containers_list:
        expected_containers_list_df = pandas.DataFrame(
            get_container_states_list(expected_containers_list),
            columns=['container_host', 'container_name', 'container_state'])

    elif os.path.exists(rhosp_containers.expected_containers_file):
        expected_containers_list_df = pandas.read_csv(
            rhosp_containers.expected_containers_file)

    failures = []
    start = time.time()
    error_info = 'Output explanation: left_only is the original state, ' \
                 'right_only is the new state'

    while time.time() - start < timeout:

        failures = []
        actual_containers_list_df = list_containers_df()

        LOG.info('expected_containers_list_df: {} '.format(
            expected_containers_list_df.to_string(index=False)))
        LOG.info('actual_containers_list_df: {} '.format(
            actual_containers_list_df.to_string(index=False)))

        # execute a `dataframe` diff between the expected and actual containers
        expected_containers_state_changed = \
            rhosp_containers.dataframe_difference(
                    expected_containers_list_df,
                    actual_containers_list_df)
        # check for changed state containerstopology
        if not expected_containers_state_changed.empty:
            failures.append('expected containers changed state ! : '
                            '\n\n{}\n{}'.format(
                             expected_containers_state_changed.
                             to_string(index=False), error_info))
            LOG.info('container states mismatched:\n{}\n'.format(failures))
            time.sleep(interval)
            # clear cache to obtain new data
            list_node_containers.cache_clear()
        else:
            LOG.info("assert_equal_containers_state :"
                     " OK, all containers are on the same state")
            return
    if failures:
        tobiko.fail(
            'container states mismatched:\n{!s}'.format('\n'.join(failures)),
            rhosp_containers.ContainerMismatchException)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools

from oslo_config import cfg


# Group with config options common for RHOSP deployments
# (both Tripleo and Podified)
RHOSP_GROUP_NAME = 'rhosp'
# Group with config options related only for Tripleo based deployments
TRIPLEO_GROUP_NAME = 'tripleo'
# Group with config options related only for podified (NextGen) deployments
PODIFIED_GROUP_NAME = "podified"

RHOSP_OPTIONS = [
    cfg.IntOpt('ssh_port',
               default=None,
               deprecated_name='overcloud_ssh_port',
               deprecated_group=TRIPLEO_GROUP_NAME,
               help="TCP port of SSH server on overcloud or EDPM compute "
                    "hosts"),
    cfg.StrOpt('ssh_username',
               default=None,
               deprecated_name='overcloud_ssh_username',
               deprecated_group=TRIPLEO_GROUP_NAME,
               help="Default username used to connect to overcloud or EDPM "
                    "compute nodes"),
    cfg.StrOpt('ssh_key_filename',
               default='~/.ssh/osp_ssh_key',
               deprecated_name='overcloud_ssh_key_filename',
               deprecated_group=TRIPLEO_GROUP_NAME,
               help="SSH key filename used to login to Overcloud or EDPM "
                    "compute nodes"),
    cfg.StrOpt('cloud_name',
               default='overcloud',
               deprecated_name='overcloud_cloud_name',
               deprecated_group=TRIPLEO_GROUP_NAME,
               help='cloud name to be used for loading credentials '),
    cfg.IntOpt('ip_version',
               deprecated_name='overcloud_ip_version',
               deprecated_group=TRIPLEO_GROUP_NAME,
               help=("Default IP address version to be used to connect to "
                     "overcloud or EDPM compute nodes")),

    cfg.StrOpt('inventory_file',
               default='.ansible/inventory/tripleo.yaml',
               deprecated_group=TRIPLEO_GROUP_NAME,
               help="path to where to export tripleo inventory file"),

    cfg.BoolOpt('has_external_load_balancer',
                default=False,
                deprecated_group=TRIPLEO_GROUP_NAME,
                help="OSP env was done with an external load balancer"),

    cfg.BoolOpt('ceph_rgw',
                default=False,
                deprecated_group=TRIPLEO_GROUP_NAME,
                help="whether Ceph RGW is deployed"),
]

TRIPLEO_OPTIONS = [
    # Undercloud options
    cfg.StrOpt('undercloud_ssh_hostname',
               default='undercloud-0',
               help="hostname or IP address to be used to connect to "
                    "undercloud host"),
    cfg.IntOpt('undercloud_ssh_port',
               default=None,
               help="TCP port of SSH server on undercloud host"),
    cfg.StrOpt('undercloud_ssh_username',
               default='stack',
               help="Username with access to stackrc and overcloudrc files"),
    cfg.StrOpt('undercloud_ssh_key_filename',
               default=None,
               help="SSH key filename used to login to Undercloud node"),
    cfg.ListOpt('undercloud_rcfile',
                default=['~/stackrc'],
                help="Undercloud RC filename"),
    cfg.StrOpt('undercloud_cloud_name',
               default='undercloud',
               help='undercloud cloud name to be used for loading credentials '
                    'from the undercloud clouds files'),
    cfg.StrOpt('undercloud_cacert_file',
               default='/etc/pki/tls/certs/ca-bundle.trust.crt',
               help='Path to cacert file that can be used to send https '
                    'request from the undercloud'),

    cfg.ListOpt('overcloud_rcfile',
                default=['~/overcloudrc', '~/qe-Cloud-0rc'],
                help="Overcloud RC filenames"),
    cfg.StrOpt('overcloud_network_name',
               help="Name of network used to connect to overcloud nodes"),
    cfg.DictOpt('overcloud_groups_dict',
                help='Dictionary with the node groups corresponding to '
                     'different hostname prefixes',
                default={'ctrl': 'controller', 'cmp': 'compute'}),
]


def register_tobiko_options(conf):
    conf.register_opts(group=cfg.OptGroup(RHOSP_GROUP_NAME),
                       opts=RHOSP_OPTIONS)
    conf.register_opts(group=cfg.OptGroup(TRIPLEO_GROUP_NAME),
                       opts=TRIPLEO_OPTIONS)


def list_options():
    return [(RHOSP_GROUP_NAME, itertools.chain(RHOSP_OPTIONS)),
            (TRIPLEO_GROUP_NAME, itertools.chain(TRIPLEO_OPTIONS))]



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import time

from oslo_log import log
import testtools

import tobiko
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.openstack import nova
from tobiko.openstack import stacks


LOG = log.getLogger(__name__)


class RebootHostStack(stacks.CirrosServerStackFixture):
    """Server to be rebooted"""


class RebootHostTest(testtools.TestCase):

    stack = tobiko.required_fixture(RebootHostStack)

    def test_reboot_host(self, nova_reboot=False, **params):
        server = self.stack.ensure_server_status('ACTIVE')
        self.assertEqual('ACTIVE', server.status)

        ssh_client = self.stack.ssh_client
        uptime_0 = sh.get_uptime(ssh_client=ssh_client)
        LOG.debug("Testing reboot command on remote host: "
                  "uptime=%r", uptime_0)
        boottime_0 = time.time() - uptime_0

        # Wait for CirrOS init script to terminate before rebooting the VM
        sh.wait_for_processes(command='^{.*}',
                              ssh_client=ssh_client,
                              timeout=90.)

        reboot = sh.reboot_host(ssh_client=ssh_client, **params)
        self.assertIs(ssh_client, reboot.ssh_client)
        self.assertEqual(ssh_client.hostname, reboot.hostname)
        method = params.get('method') or sh.soft_reset_method
        self.assertIs(method, reboot.method)

        if not reboot.is_rebooted:
            self.assert_is_not_connected(ssh_client)
            if nova_reboot:
                ping.ping_until_unreceived(self.stack.ip_address)
                nova.reboot_server(server)
            reboot.wait_for_operation()

        self.assertTrue(reboot.is_rebooted)
        self.assert_is_connected(ssh_client)

        server = nova.wait_for_server_status(self.stack.server_id, 'ACTIVE')
        self.assertEqual('ACTIVE', server.status)

        uptime_1 = sh.get_uptime(ssh_client=ssh_client)
        boottime_1 = time.time() - uptime_1
        LOG.debug("Reboot operation executed on remote host: "
                  "uptime=%r", uptime_1)
        self.assertGreater(boottime_1, boottime_0)

    def test_reboot_host_with_chash_method(self):
        self.test_reboot_host(method=sh.crash_method,
                              wait=False,
                              nova_reboot=True)

    def test_reboot_host_with_hard_method(self):
        self.test_reboot_host(method=sh.hard_reset_method)

    def test_reboot_host_with_soft_method(self):
        self.test_reboot_host(method=sh.soft_reset_method)

    def test_reboot_host_with_invalid_method(self):
        self.assertRaises(TypeError,
                          sh.reboot_host,
                          ssh_client=self.stack.ssh_client,
                          method='<invalid-method>')

    def test_reboot_host_with_wait(self):
        self.test_reboot_host(wait=True)

    def test_reboot_host_with_no_wait(self):
        self.test_reboot_host(wait=False)

    def test_reboot_server_when_shutoff(self):
        server = self.stack.ensure_server_status('SHUTOFF')
        self.assertEqual('SHUTOFF', server.status)

        ssh_client = self.stack.ssh_client
        self.assert_is_not_connected(ssh_client)
        self.assertRaises(sh.RebootHostTimeoutError,
                          sh.reboot_host,
                          ssh_client=ssh_client,
                          timeout=5.0)
        self.assert_is_not_connected(ssh_client)
        server = nova.get_server(self.stack.server_id)
        self.assertEqual('SHUTOFF', server.status)

    def assert_is_connected(self, ssh_client):
        self.assertIsNotNone(ssh_client.client)

    def assert_is_not_connected(self, ssh_client):
        self.assertIsNone(ssh_client.client)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools
import json
import os
import typing

from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko import rhosp
from tobiko.shell import ssh
from tobiko.shell import sh


CONF = config.CONF

LOG = log.getLogger(__name__)


def undercloud_ssh_client() -> ssh.SSHClientFixture:
    host_config = undercloud_host_config()
    if not host_config.hostname:
        raise NoSuchUndercloudHostname('No such undercloud hostname')
    return ssh.ssh_client(host=host_config.hostname,
                          **host_config.connect_parameters)


class NoSuchUndercloudHostname(tobiko.TobikoException):
    message = "Undercloud hostname not specified"


class InvalidRCFile(tobiko.TobikoException):
    message = "Invalid RC file: {rcfile}"


@functools.lru_cache()
def fetch_os_env(rcfile: str, *rcfiles: str) -> typing.Dict[str, str]:
    rcfiles = (rcfile,) + rcfiles
    LOG.debug('Fetching OS environment variables from TripleO undercloud '
              f'host files: {",".join(rcfiles)}')
    errors = []
    for rcfile in rcfiles:
        LOG.debug(f'Reading rcfile: {rcfile}...')
        try:
            result = sh.execute(f". {rcfile}; env | grep '^OS_'",
                                ssh_client=undercloud_ssh_client())
        except sh.ShellCommandFailed as ex:
            LOG.debug(f"Unable to get overcloud RC file '{rcfile}' content "
                      f"({ex})")
            errors.append(ex)
        else:
            LOG.debug(f'Parsing environment variables from: {rcfile}...')
            env = {}
            for line in result.stdout.splitlines():
                name, value = line.split('=')
                env[name] = value
            if env:
                env_dump = json.dumps(env, sort_keys=True, indent=4)
                LOG.debug(f'Environment variables read from: {rcfile}:\n'
                          f'{env_dump}')
                return env
    if errors and all("No such file or directory" in error.stderr
                      for error in errors):
        LOG.warning('None of the credentials files were found')
        raise keystone.NoSuchKeystoneCredentials()
    raise InvalidRCFile(rcfile=", ".join(rcfiles))


def load_undercloud_rcfile() -> typing.Dict[str, str]:
    conf = tobiko.tobiko_config().tripleo
    return fetch_os_env(*conf.undercloud_rcfile)


class UndercloudKeystoneCredentialsFixtureBase(
        keystone.KeystoneCredentialsFixture):

    def _get_credentials(self) -> keystone.KeystoneCredentials:
        if not has_undercloud():
            raise keystone.NoSuchKeystoneCredentials()
        return super()._get_credentials()

    def _get_connection(self) -> sh.ShellConnectionType:
        return undercloud_ssh_client()

    def _get_environ(self) -> typing.Dict[str, str]:
        return load_undercloud_rcfile()


class UndercloudCloudsFileKeystoneCredentialsFixture(
        UndercloudKeystoneCredentialsFixtureBase,
        keystone.CloudsFileKeystoneCredentialsFixture):

    @staticmethod
    def _get_default_cloud_name() -> typing.Optional[str]:
        return tobiko.tobiko_config().tripleo.undercloud_cloud_name


class UndercloudEnvironKeystoneCredentialsFixture(
        UndercloudKeystoneCredentialsFixtureBase,
        keystone.EnvironKeystoneCredentialsFixture):
    pass


@functools.lru_cache()
def has_undercloud(min_version: tobiko.VersionType = None,
                   max_version: tobiko.VersionType = None) -> bool:
    try:
        check_undercloud(min_version=min_version,
                         max_version=max_version)
    except (UndercloudNotFound, UndercloudVersionMismatch) as ex:
        LOG.debug(f'TripleO undercloud host not found:\n'
                  f'{ex}')
        return False
    except Exception:
        LOG.exception('Error looking for undercloud host')
        return False
    else:
        LOG.debug('TripleO undercloud host found')
        return True


skip_if_missing_undercloud = tobiko.skip_unless(
    'TripleO undercloud hostname not configured', has_undercloud)


def skip_unlsess_has_undercloud(min_version: tobiko.VersionType = None,
                                max_version: tobiko.VersionType = None):
    return tobiko.skip_on_error(
        reason='TripleO undercloud not found',
        predicate=check_undercloud,
        min_version=min_version,
        max_version=max_version,
        error_type=(UndercloudNotFound, UndercloudVersionMismatch))


class UndecloudHostConfig(tobiko.SharedFixture):

    hostname: typing.Optional[str] = None
    port: typing.Optional[int] = None
    username: typing.Optional[str] = None

    def __init__(self, **kwargs):
        super(UndecloudHostConfig, self).__init__()
        self._connect_parameters = ssh.gather_ssh_connect_parameters(**kwargs)

    def setup_fixture(self):
        self.hostname = CONF.tobiko.tripleo.undercloud_ssh_hostname.strip()
        self.port = CONF.tobiko.tripleo.undercloud_ssh_port
        self.username = CONF.tobiko.tripleo.undercloud_ssh_username

    @property
    def key_filename(self) -> typing.List[str]:
        key_filenames: typing.List[str] = []
        conf = tobiko.tobiko_config()
        key_filename = conf.tripleo.undercloud_ssh_key_filename
        if key_filename:
            key_filename = tobiko.tobiko_config_path(key_filename)
            if os.path.isfile(key_filename):
                key_filenames.append(key_filename)
        key_filenames.extend(ssh.list_proxy_jump_key_filenames())
        key_filenames.extend(ssh.list_key_filenames())
        return tobiko.select_uniques(key_filenames)

    @property
    def connect_parameters(self):
        parameters = ssh.gather_ssh_connect_parameters(self)
        parameters.update(self._connect_parameters)
        return parameters


def undercloud_host_config() -> UndecloudHostConfig:
    return tobiko.setup_fixture(UndecloudHostConfig)


def undercloud_keystone_client():
    session = undercloud_keystone_session()
    return keystone.get_keystone_client(session=session)


class UndercloudKeystoneCredentialsFixture(
        UndercloudKeystoneCredentialsFixtureBase,
        keystone.DelegateKeystoneCredentialsFixture):

    @staticmethod
    def _get_delegates() -> typing.List[keystone.KeystoneCredentialsFixture]:
        return [
            tobiko.get_fixture(
                UndercloudCloudsFileKeystoneCredentialsFixture),
            tobiko.get_fixture(
                UndercloudEnvironKeystoneCredentialsFixture)]


def undercloud_keystone_session() -> keystone.KeystoneSession:
    credentials = undercloud_keystone_credentials()
    return keystone.get_keystone_session(credentials=credentials)


def undercloud_keystone_credentials() -> keystone.KeystoneCredentialsFixture:
    return tobiko.get_fixture(UndercloudKeystoneCredentialsFixture)


@functools.lru_cache()
def undercloud_version() -> tobiko.Version:
    ssh_client = undercloud_ssh_client()
    return rhosp.get_rhosp_version(connection=ssh_client)


def check_undercloud(min_version: tobiko.Version = None,
                     max_version: tobiko.Version = None):
    try:
        ssh_client = undercloud_ssh_client()
    except NoSuchUndercloudHostname as ex:
        raise UndercloudNotFound(
            cause='TripleO undercloud hostname not found') from ex
    try:
        ssh_client.connect(retry_count=1,
                           connection_attempts=1,
                           timeout=15.)
    except Exception as ex:
        raise UndercloudNotFound(
            cause=f'unable to connect to TripleO undercloud host: {ex}'
        ) from ex

    if min_version or max_version:
        tobiko.check_version(undercloud_version(),
                             min_version=min_version,
                             max_version=max_version,
                             mismatch_error=UndercloudVersionMismatch)


class UndercloudNotFound(tobiko.ObjectNotFound):
    message = 'undercloud not found: {cause}'


class UndercloudVersionMismatch(tobiko.VersionMismatch):
    message = 'undercloud version mismatch: {version} {cause}'



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from unittest import mock

from tobiko.tests import unit
from tobiko.openstack import nova


class ServerTest(unit.TobikoUnitTest):

    def test_get_server_id_with_str(self):
        self.assertEqual('<server-id>',
                         nova.get_server_id('<server-id>'))

    def test_get_server_id_with_server(self):
        server_mock = mock.Mock(id='<server-id>')
        self.assertEqual('<server-id>',
                         nova.get_server_id(server_mock))



# Copyright 2023 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.manila import _client
from tobiko.openstack.manila import _constants
from tobiko.openstack.manila import _exceptions
from tobiko.openstack.manila import _waiters

manila_client = _client.manila_client
get_manila_client = _client.get_manila_client
ManilaClientFixture = _client.ManilaClientFixture
create_share = _client.create_share
get_share = _client.get_share
get_shares_by_name = _client.get_shares_by_name
delete_share = _client.delete_share
extend_share = _client.extend_share
list_shares = _client.list_shares
ensure_default_share_type_exists = _client.ensure_default_share_type_exists

# Waiters
wait_for_share_status = _waiters.wait_for_share_status
wait_for_resource_deletion = _waiters.wait_for_resource_deletion

# Exceptions
ShareNotFound = _exceptions.ShareNotFound
ShareReleaseFailed = _exceptions.ShareReleaseFailed

# Constants
RESOURCE_STATUS = _constants.RESOURCE_STATUS
STATUS_AVAILABLE = _constants.STATUS_AVAILABLE
STATUS_ERROR = _constants.STATUS_ERROR
STATUS_ERROR_DELETING = _constants.STATUS_ERROR_DELETING
SHARE_NAME = _constants.SHARE_NAME



# Copyright (c) 2023 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko.openstack import topology as osp_topology
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.tests.functional.openstack import test_topology
from tobiko import podified


@podified.skip_if_not_podified
class PodifiedTopologyTest(test_topology.OpenStackTopologyTest):

    expected_group: osp_topology.OpenstackGroupNamesType = 'compute'

    @property
    def topology(self) -> podified.PodifiedTopology:
        return tobiko.setup_fixture(podified.PodifiedTopology)

    def test_ping_node(self):
        # NOTE(slaweq): in podified topology we expect connectivity only to the
        # edpm nodes, not to the OCP workers
        for node in self.topology.get_group("compute"):
            ping.ping(node.public_ip, count=1, timeout=5.).assert_replied()

    def test_ssh_client(self):
        # NOTE(slaweq): in podified topology we expect connectivity only to the
        # edpm nodes, not to the OCP workers
        for node in self.topology.get_group("compute"):
            self.assertIsNotNone(node.ssh_client)
            hostname = sh.ssh_hostname(
                ssh_client=node.ssh_client).split('.')[0]
            self.assertEqual(node.name, hostname)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import contextlib
import typing

import novaclient
import novaclient.exceptions
import novaclient.v2.client
import novaclient.v2.servers
import novaclient.v2.hypervisors
from oslo_log import log

import tobiko
from tobiko.openstack import _client


LOG = log.getLogger(__name__)

CLIENT_CLASSES = (novaclient.v2.client.Client,)
NovaClient = typing.Union[novaclient.v2.client.Client]
NovaServer = typing.Union[novaclient.v2.servers.Server]
NovaHypervisor = typing.Union[novaclient.v2.hypervisors.Hypervisor]


class NovaClientFixture(_client.OpenstackClientFixture):

    def init_client(self, session) -> NovaClient:
        return novaclient.client.Client('2.56', session=session)


class NovaClientManager(_client.OpenstackClientManager):

    def create_client(self, session) -> NovaClientFixture:
        return NovaClientFixture(session=session)


CLIENTS = NovaClientManager()

NovaClientType = typing.Union[NovaClient,
                              NovaClientFixture,
                              typing.Type[NovaClientFixture],
                              None]


def nova_client(obj: NovaClientType) -> NovaClient:
    if obj is None:
        return get_nova_client()

    if isinstance(obj, CLIENT_CLASSES):
        return obj

    fixture = tobiko.setup_fixture(obj)
    if isinstance(fixture, NovaClientFixture):
        assert fixture.client is not None
        return fixture.client

    message = f"Object '{obj}' is not a NovaClientFixture"
    raise TypeError(message)


def get_nova_client(session=None, shared=True, init_client=None,
                    manager=None) -> NovaClient:
    manager = manager or CLIENTS
    client = manager.get_client(session=session, shared=shared,
                                init_client=init_client)
    tobiko.setup_fixture(client)
    return client.client


def list_hypervisors(client: NovaClientType = None, detailed=True, **params) \
        -> tobiko.Selection[NovaHypervisor]:
    client = nova_client(client)
    hypervisors = client.hypervisors.list(detailed=detailed)
    return tobiko.select(hypervisors).with_attributes(**params)


def find_hypervisor(client: NovaClientType = None, unique=False, **params) \
        -> NovaHypervisor:
    hypervisors = list_hypervisors(client=client, **params)
    if unique:
        return hypervisors.unique
    else:
        return hypervisors.first


def list_servers(client: NovaClientType = None, **params) -> \
        tobiko.Selection[NovaServer]:
    servers = nova_client(client).servers.list()
    return tobiko.select(servers).with_attributes(**params)


def find_server(client: NovaClientType = None, unique=False, **params) -> \
        NovaServer:
    servers = list_servers(client=client, **params)
    if unique:
        return servers.unique
    else:
        return servers.first


def list_services(client: NovaClientType = None, **params) -> tobiko.Selection:
    client = nova_client(client)
    services = client.services.list()
    return tobiko.select(services).with_attributes(**params)


def find_service(client: NovaClientType = None, unique=False, **params):
    services = list_services(client=client, **params)
    if unique:
        return services.unique
    else:
        return services.first


ServerType = typing.Union[str, NovaServer]


def get_server_id(server: ServerType = None,
                  server_id: str = None) -> str:
    if server_id is None:
        if isinstance(server, str):
            server_id = server
        else:
            assert server is not None
            server_id = server.id
    return server_id


class GetServerError(tobiko.TobikoException):
    message = ("Error getting server '{server_id}' details "
               "(params={params}):\n"
               "{reason}")


class ServerNotFoundError(GetServerError, tobiko.ObjectNotFound):
    pass


def get_server(server: ServerType = None,
               server_id: str = None,
               client: NovaClientType = None,
               **params) -> NovaServer:
    server_id = get_server_id(server=server, server_id=server_id)
    try:
        return nova_client(client).servers.get(server_id, **params)
    except novaclient.exceptions.NotFound as ex:
        raise ServerNotFoundError(server_id=server_id,
                                  params=params,
                                  reason=str(ex)) from ex


def delete_server(server: ServerType = None,
                  server_id: str = None,
                  client: NovaClientType = None,
                  **params):
    server_id = get_server_id(server=server, server_id=server_id)
    try:
        return nova_client(client).servers.delete(server_id, **params)
    except novaclient.exceptions.NotFound as ex:
        raise ServerNotFoundError(server_id=server_id,
                                  params=params,
                                  reason=str(ex)) from ex


def migrate_server(server: ServerType = None,
                   server_id: str = None,
                   host: str = None,
                   client: NovaClientType = None,
                   **params):
    server_id = get_server_id(server=server, server_id=server_id)
    if host is not None:
        params.update(host=host)
    LOG.debug(f"Start server '{server_id}' migration...\n" +
              f"{params}")
    with handle_migration_errors(server_id=server_id, **params):
        # pylint: disable=protected-access
        return nova_client(client).servers._action(
            'migrate', server_id, info=params)


def live_migrate_server(server: ServerType = None,
                        server_id: str = None,
                        host: str = None,
                        block_migration: bool = None,
                        client: NovaClientType = None,
                        **params):
    server_id = get_server_id(server=server, server_id=server_id)
    params.update(host=host)
    if block_migration is None:
        params.update(block_migration='auto')
    else:
        params.update(block_migration=block_migration)

    LOG.debug(f"Start server '{server_id}' live migration...\n" +
              f"{params}")
    with handle_migration_errors(server_id=server_id, **params):
        return nova_client(client).servers.live_migrate(server=server_id,
                                                        **params)


@contextlib.contextmanager
def handle_migration_errors(server_id: str, **params):
    try:
        yield
    except novaclient.exceptions.BadRequest as ex:
        reason = str(ex)
        error_class = MigrateServerError
        if 'No valid host was found' in reason:
            error_class = NoValidHostFoundMigrateServerError
        if 'Migration pre-check error' in reason:
            error_class = PreCheckMigrateServerError
        if 'is not on shared storage' in reason:
            error_class = NotInSharedStorageMigrateServerError
        if 'is not on local storage' in reason:
            error_class = NotInLocalStorageMigrateServerError
        raise error_class(server_id=server_id,
                          reason=reason,
                          params=params) from ex


class MigrateServerError(tobiko.TobikoException):
    message = ("Error migrating server '{server_id}' "
               "(params={params}):\n"
               "{reason}")


class NoValidHostFoundMigrateServerError(MigrateServerError):
    pass


class PreCheckMigrateServerError(MigrateServerError):
    pass


class NotInLocalStorageMigrateServerError(MigrateServerError):
    pass


class NotInSharedStorageMigrateServerError(MigrateServerError):
    pass


def confirm_resize(server: typing.Optional[ServerType] = None,
                   server_id: typing.Optional[str] = None,
                   client: NovaClientType = None, **params):
    server_id = get_server_id(server=server, server_id=server_id)
    LOG.debug(f"Confirm server resize (server_id='{server_id}', "
              f"info={params})")
    return nova_client(client).servers.confirm_resize(server_id, **params)


MAX_SERVER_CONSOLE_OUTPUT_LENGTH = 1024 * 256


def get_console_output(server: typing.Optional[ServerType] = None,
                       server_id: typing.Optional[str] = None,
                       timeout: tobiko.Seconds = None,
                       interval: tobiko.Seconds = None,
                       length: typing.Optional[int] = None,
                       client: NovaClientType = None) -> \
        typing.Optional[str]:
    if length is not None:
        length = min(length, MAX_SERVER_CONSOLE_OUTPUT_LENGTH)
    else:
        length = MAX_SERVER_CONSOLE_OUTPUT_LENGTH

    server_id = get_server_id(server=server, server_id=server_id)

    for attempt in tobiko.retry(timeout=timeout,
                                interval=interval,
                                default_timeout=60.,
                                default_interval=5.):
        try:
            output = nova_client(client).servers.get_console_output(
                server=server_id, length=length)
        except (TypeError, novaclient.exceptions.NotFound):
            # Only active servers have console output
            server = get_server(server_id=server_id)
            if server.status != 'ACTIVE':
                LOG.debug(f"Server '{server_id}' has no console output "
                          f"(status = '{server.status}').")
                break
            else:
                # For some reason it could happen resulting body cannot be
                # translated to json object and it is converted to None
                LOG.exception(f"Error getting server '{server_id}' console "
                              "output")
        else:
            if output:
                LOG.debug(f"got server '{server_id}' console output "
                          f"(length = {len(output)}).")
                return output

        if attempt.is_last:
            LOG.info(f"No console output produced by server '{server_id}') "
                     f" after {attempt.elapsed_time} seconds")
            break

        LOG.debug(f"Waiting for server '{server_id}' console output...")

    return None


def wait_for_guest_os_ready(server: typing.Optional[ServerType] = None,
                            server_id: typing.Optional[str] = None,
                            timeout: tobiko.Seconds = 120.,
                            interval: tobiko.Seconds = 5.,
                            client: NovaClientType = None) -> None:

    def system_booted():
        console_output = get_console_output(
            server=server, server_id=server_id, client=client) or ''
        for line in console_output.splitlines():
            if 'login:' in line.lower():
                return True
        return False

    for _ in tobiko.retry(timeout=timeout, interval=interval):
        if system_booted():
            LOG.debug('VM boot completed successfully')
            return
        else:
            LOG.debug('VM boot not completed yet')


class HasNovaClientMixin(object):

    nova_client: NovaClientType = None

    def get_server(self, server: ServerType, **params) -> NovaServer:
        return get_server(server=server, client=self.nova_client, **params)

    def get_server_console_output(self, server: ServerType, **params) -> \
            typing.Optional[str]:
        return get_console_output(server=server, client=self.nova_client,
                                  **params)


class WaitForServerStatusError(tobiko.TobikoException):
    message = ("Server {server_id} not changing status from {server_status} "
               "to {status}")


class WaitForServerStatusTimeout(WaitForServerStatusError):
    message = ("Server {server_id} didn't change its status from "
               "{server_status} to {status} status after {timeout} seconds")


NOVA_SERVER_TRANSIENT_STATUS: typing.Dict[str, typing.Set[str]] = {
    'ACTIVE': {'BUILD', 'SHUTOFF', 'REBOOT'},
    'SHUTOFF': {'ACTIVE'},
    'VERIFY_RESIZE': {'RESIZE'},
}


def wait_for_server_status(
        server: ServerType,
        status: str,
        client: NovaClientType = None,
        timeout: tobiko.Seconds = None,
        sleep_time: tobiko.Seconds = None,
        transient_status: typing.Optional[typing.Container[str]] = None) -> \
            NovaServer:
    if transient_status is None:
        transient_status = NOVA_SERVER_TRANSIENT_STATUS.get(status) or []
    server_id = get_server_id(server)
    for attempt in tobiko.retry(timeout=timeout,
                                interval=sleep_time,
                                default_timeout=300.,
                                default_interval=5.):
        _server = get_server(server_id=server_id, client=client)
        if _server.status == status:
            break

        if _server.status not in transient_status:
            raise WaitForServerStatusError(server_id=server_id,
                                           server_status=_server.status,
                                           status=status)
        try:
            attempt.check_time_left()
        except tobiko.RetryTimeLimitError as ex:
            raise WaitForServerStatusTimeout(server_id=server_id,
                                             server_status=_server.status,
                                             status=status,
                                             timeout=timeout) from ex

        progress = getattr(server, 'progress', None)
        LOG.debug(f"Waiting for server {server_id} status to get from "
                  f"{_server.status} to {status} "
                  f"(progress={progress}%)")
    else:
        raise RuntimeError("Broken retry loop")

    return _server


def shutoff_server(server: ServerType = None,
                   client: NovaClientType = None,
                   timeout: tobiko.Seconds = None,
                   sleep_time: tobiko.Seconds = None) -> NovaServer:
    client = nova_client(client)
    server = get_server(server=server, client=client)
    if server.status == 'SHUTOFF':
        return server

    LOG.info(f"stop server '{server.id}' (status='{server.status}').")
    client.servers.stop(server.id)
    return wait_for_server_status(server=server.id,
                                  status='SHUTOFF',
                                  client=client,
                                  timeout=timeout,
                                  sleep_time=sleep_time)


def activate_server(server: ServerType,
                    client: NovaClientType = None,
                    timeout: tobiko.Seconds = None,
                    sleep_time: tobiko.Seconds = None) -> NovaServer:
    client = nova_client(client)
    server = get_server(server=server, client=client)
    if server.status == 'ACTIVE':
        return server

    if server.status == 'SHUTOFF':
        LOG.info(f"Start server '{server.id}' (status='{server.status}').")
        client.servers.start(server.id)
    elif server.status == 'RESIZE':
        server = wait_for_server_status(
            server=server.id, status='VERIFY_RESIZE', client=client,
            timeout=timeout, sleep_time=sleep_time)
        LOG.info(f"Confirm resize of server '{server.id}' "
                 f"(status='{server.status}').")
        client.servers.confirm_resize(server)
    elif server.status == 'VERIFY_RESIZE':
        LOG.info(f"Confirm resize of server '{server.id}' "
                 f"(status='{server.status}').")
        client.servers.confirm_resize(server)
    elif server.status != 'REBOOT':
        LOG.warning(f"Try activating server '{server.id}' by rebooting "
                    f"it  (status='{server.status}').")
        client.servers.reboot(server.id, reboot_type='HARD')

    return wait_for_server_status(server=server.id, status='ACTIVE',
                                  client=client,
                                  timeout=timeout,
                                  sleep_time=sleep_time)


def reboot_server(server: ServerType,
                  client: NovaClientType = None,
                  timeout: tobiko.Seconds = None,
                  sleep_time: tobiko.Seconds = None) -> NovaServer:
    client = nova_client(client)
    server = get_server(server=server, client=client)
    if server.status == 'REBOOT':
        return server

    if server.status == 'SHUTOFF':
        LOG.info(f"Start server '{server.id}' (status='{server.status}').")
        client.servers.start(server.id)
    else:
        LOG.info(f"Reboot server '{server.id}' (status='{server.status}').")
        client.servers.reboot(server.id)

    return wait_for_server_status(server=server.id, status='ACTIVE',
                                  client=client, timeout=timeout,
                                  sleep_time=sleep_time)


def ensure_server_status(server: ServerType,
                         status: str,
                         client: NovaClientType = None,
                         timeout: tobiko.Seconds = None,
                         sleep_time: tobiko.Seconds = None) -> NovaServer:
    if status == 'ACTIVE':
        return activate_server(server=server, client=client, timeout=timeout,
                               sleep_time=sleep_time)
    elif status == 'SHUTOFF':
        return shutoff_server(server=server, client=client, timeout=timeout,
                              sleep_time=sleep_time)
    else:
        raise ValueError(f"Unsupported server status: '{status}'")



# Copyright (c) 2023 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools
from oslo_log import log

from tobiko import config
from tobiko.openstack import keystone
from tobiko.openstack import manila

LOG = log.getLogger(__name__)
CONF = config.CONF


@keystone.skip_if_missing_service(name='manila')
class ManilaApiTestCase(testtools.TestCase):
    """Manila scenario tests.

    Create a manila share.
    Check it reaches status 'available'.
    After upgrade/disruptions/etc, check the share is still valid and it can be
    extended.
    """
    @classmethod
    def setUpClass(cls):
        if config.get_bool_env('TOBIKO_PREVENT_CREATE'):
            LOG.debug('skipping creation of manila resources')
            cls.share = manila.get_shares_by_name(manila.SHARE_NAME)[0]
        else:
            manila.ensure_default_share_type_exists()
            cls.share = manila.create_share(name=manila.SHARE_NAME)

        manila.wait_for_share_status(cls.share['id'])

    @config.skip_if_prevent_create()
    def test_1_create_share(self):
        self.assertEqual(manila.SHARE_NAME, self.share['name'])

    @config.skip_unless_prevent_create()
    def test_2_extend_share(self):
        share_id = self.share['id']
        manila.extend_share(share_id, new_size=CONF.tobiko.manila.size + 1)
        manila.wait_for_share_status(share_id)
        share_size = manila.get_share(share_id)['size']
        self.assertEqual(CONF.tobiko.manila.size + 1, share_size)



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import importlib
import itertools
import logging
import os
import typing  # noqa

from oslo_config import cfg
from oslo_log import log

import tobiko


LOG = log.getLogger(__name__)


CONFIG_MODULES = ['tobiko.common._case',
                  'tobiko.openstack.glance.config',
                  'tobiko.openstack.manila.config',
                  'tobiko.openstack.keystone.config',
                  'tobiko.openstack.neutron.config',
                  'tobiko.openstack.nova.config',
                  'tobiko.openstack.octavia.config',
                  'tobiko.openstack.topology.config',
                  'tobiko.shell.ansible.config',
                  'tobiko.shell.ssh.config',
                  'tobiko.shell.ping.config',
                  'tobiko.shell.iperf3.config',
                  'tobiko.shell.sh.config',
                  'tobiko.shiftstack.config',
                  'tobiko.rhosp.config',
                  'tobiko.podified.config',
                  'tobiko.tripleo.config']


LOGGING_CONF_GROUP_NAME = "logging"

LOGGING_OPTIONS = [
    cfg.BoolOpt('capture_log',
                default=True,
                help="Whenever to report debugging log lines"),
    cfg.StrOpt('line_format',
               default=('%(asctime)s.%(msecs)03d %(process)d %(levelname)s '
                        '%(name)s - %(message)s'),
               help='Default logging line format string'),
    cfg.StrOpt('date_format',
               default='%Y-%m-%d %H:%M:%S',
               help='Default logging date format string'),
]

HTTP_CONF_GROUP_NAME = "http"

HTTP_OPTIONS = [
    cfg.StrOpt('http_proxy',
               help="HTTP proxy URL for Rest APIs"),
    cfg.StrOpt('https_proxy',
               help="HTTPS proxy URL for Rest APIs"),
    cfg.StrOpt('no_proxy',
               help="Don't use proxy server to connect to listed hosts")]

TESTCASE_CONF_GROUP_NAME = "testcase"

TESTCASE_OPTIONS = [
    cfg.FloatOpt('timeout',
                 default=None,
                 help=("Timeout (in seconds) used for interrupting test case "
                       "execution")),
    cfg.FloatOpt('test_runner_timeout',
                 default=None,
                 help=("Timeout (in seconds) used for interrupting test "
                       "runner execution"))]

COMMON_GROUP_NAME = 'common'

COMMON_OPTIONS = [
    cfg.StrOpt('shelves_dir',
               default='~/.tobiko/cache/shelves',
               help="Directory where to look for shelves."),
    cfg.StrOpt('lock_dir',
               default='~/.tobiko/cache/lock',
               help="Directory where lock persistent files will be saved"),
]


def workspace_config_files(project=None, prog=None):
    project = project or 'tobiko'
    filenames = []
    if prog is not None:
        filenames.append(prog + '.conf')
    filenames.append(project + '.conf')
    root_dir = os.path.realpath("/")
    current_dir = os.path.realpath(os.getcwd())
    config_files = []
    while current_dir != root_dir:
        for filename in filenames:
            filename = os.path.join(current_dir, filename)
            if os.path.isfile(filename):
                config_files.append(filename)
        current_dir = os.path.dirname(current_dir)
    return config_files


class GlobalConfig(object):

    # this is a singletone
    _instance = None
    _sources = {}  # type: typing.Dict[str, typing.Any]

    def __new__(cls):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
        return cls._instance

    def set_source(self, name, source_conf):
        if source_conf is None:
            raise TypeError("Config source cannot be None")
        actual = self._sources.setdefault(name, source_conf)
        if actual is not source_conf:
            msg = "Config source already registered: {!r}".format(name)
            raise RuntimeError(msg)

    def __getattr__(self, name):
        sources = self._sources.get(name)
        if sources is None:
            msg = "Config source not registered: {!r}".format(name)
            raise NoSuchConfigSource(msg)
        return sources


CONF = GlobalConfig()


class InitConfigFixture(tobiko.SharedFixture):

    def setup_fixture(self):
        init_tobiko_config()
        init_environ_config()


def init_config():
    tobiko.setup_fixture(InitConfigFixture)


def get_version():
    from tobiko import version
    return version.release


def init_tobiko_config(default_config_dirs=None, default_config_files=None,
                       project=None, prog=None, product_name=None,
                       version=None):

    if project is None:
        project = 'tobiko'

    if product_name is None:
        product_name = 'tobiko'

    if version is None:
        version = get_version()

    if default_config_dirs is None:
        default_config_dirs = cfg.find_config_dirs(project=project, prog=prog)
    if default_config_files is None:
        default_config_files = (cfg.find_config_files(project=project,
                                                      prog=prog) +
                                workspace_config_files(project=project,
                                                       prog=prog))

    # Register configuration options
    conf = cfg.ConfigOpts()
    log.register_options(conf)
    register_tobiko_options(conf=conf)

    # Initialize Tobiko configuration object
    conf(args=[],
         project='tobiko',
         validate_default_values=True,
         default_config_dirs=default_config_dirs,
         default_config_files=default_config_files)
    CONF.set_source('tobiko', conf)

    # expand and normalize log_file and log_dir names
    conf.config_dir = os.path.realpath(conf.find_file('.'))
    log_dir = conf.log_dir or conf.config_dir
    log_file = conf.log_file or 'tobiko.log'
    log_path = os.path.realpath(os.path.expanduser(
        os.path.join(log_dir, log_file)))
    conf.log_dir = os.path.dirname(log_path)
    conf.log_file = os.path.basename(log_path)

    # setup final configuration
    log.setup(conf=conf, product_name=product_name, version=version)
    setup_tobiko_config(conf=conf)
    LOG.debug("Configuration setup using parameters:\n"
              " - product_name: %r\n"
              " - version: %r\n"
              " - default_config_dirs: %r\n"
              " - default_config_files: %r\n"
              " - log_file: %r\n",
              product_name,
              version,
              default_config_dirs,
              default_config_files,
              os.path.join(log_dir, log_file))


def register_tobiko_options(conf):

    conf.register_opts(
        group=cfg.OptGroup(LOGGING_CONF_GROUP_NAME), opts=LOGGING_OPTIONS)

    conf.register_opts(
        group=cfg.OptGroup(HTTP_CONF_GROUP_NAME), opts=HTTP_OPTIONS)

    conf.register_opts(
        group=cfg.OptGroup(TESTCASE_CONF_GROUP_NAME), opts=TESTCASE_OPTIONS)

    conf.register_opts(
        group=cfg.OptGroup(COMMON_GROUP_NAME), opts=COMMON_OPTIONS)

    for module_name in CONFIG_MODULES:
        module = importlib.import_module(module_name)
        if hasattr(module, 'register_tobiko_options'):
            module.register_tobiko_options(conf=conf)


def list_http_options():
    return [
        (HTTP_CONF_GROUP_NAME, itertools.chain(HTTP_OPTIONS))
    ]


def list_testcase_options():
    return [
        (TESTCASE_CONF_GROUP_NAME, itertools.chain(TESTCASE_OPTIONS))
    ]


def list_common_options():
    return [
        (COMMON_GROUP_NAME, itertools.chain(COMMON_OPTIONS))
    ]


def list_tobiko_options():
    all_options = (list_http_options() +
                   list_testcase_options() +
                   list_common_options())

    for module_name in CONFIG_MODULES:
        module = importlib.import_module(module_name)
        if hasattr(module, 'list_options'):
            all_options += module.list_options()
    return all_options


def setup_tobiko_config(conf):
    # Redirect all warnings to logging library
    logging.captureWarnings(True)
    warnings_logger = log.getLogger('py.warnings')
    if conf.debug:
        if not warnings_logger.isEnabledFor(log.WARNING):
            # Print Python warnings
            warnings_logger.logger.setLevel(log.WARNING)
    elif warnings_logger.isEnabledFor(log.WARNING):
        # Silence Python warnings
        warnings_logger.logger.setLevel(log.ERROR)

    tobiko.setup_fixture(HttpProxyFixture)

    for module_name in CONFIG_MODULES:
        module = importlib.import_module(module_name)
        if hasattr(module, 'setup_tobiko_config'):
            module.setup_tobiko_config(conf=conf)


class HttpProxyFixture(tobiko.SharedFixture):
    """Make sure we have http proxy environment variables defined when required
    """

    http_proxy = None
    https_proxy = None
    no_proxy = None
    source = None

    def setup_fixture(self):
        source = None
        http_proxy = os.environ.get('http_proxy')
        https_proxy = os.environ.get('https_proxy')
        no_proxy = os.environ.get('no_proxy')
        if http_proxy or https_proxy:
            source = 'environment'
        else:
            http_conf = CONF.tobiko.http
            http_proxy = http_conf.http_proxy
            https_proxy = http_conf.https_proxy
            no_proxy = http_conf.no_proxy
            if http_proxy:
                os.environ['http_proxy'] = http_proxy
            if https_proxy:
                os.environ['https_proxy'] = https_proxy
            if no_proxy:
                os.environ['no_proxy'] = no_proxy
            if http_proxy or https_proxy or no_proxy:
                source = 'tobiko.conf'

        if source:
            LOG.info("Using HTTP proxy configuration defined in %s:\n"
                     "  http_proxy: %r\n"
                     "  https_proxy: %r\n"
                     "  no_proxy: %r",
                     source, os.environ.get('http_proxy'),
                     os.environ.get('https_proxy'), os.environ.get('no_proxy'))
        else:
            LOG.debug("Connecting to REST API services without a proxy "
                      "server")

        self.source = source
        self.http_proxy = os.environ.get('http_proxy')
        self.https_proxy = os.environ.get('https_proxy')
        self.no_proxy = os.environ.get('no_proxy')


def init_environ_config():
    CONF.set_source('environ', EnvironConfig())


class EnvironConfig(object):

    def __getattr__(self, name):
        value = get_env(name)
        if value is None:
            msg = "Environment variable not defined: {!r}".format(name)
            raise cfg.NoSuchOptError(msg)
        return value


class NoSuchConfigSource(AttributeError):
    pass


def get_any_option(*sources, **kwargs):
    default = kwargs.get('default', None)
    for source in sources:
        value = CONF
        for name in source.split('.'):
            try:
                value = getattr(value, name)
            except (NoSuchConfigSource,
                    cfg.NoSuchOptError,
                    cfg.NoSuchGroupError) as ex:
                LOG.debug("No such option value for %r: %s", source, ex)
                break
        else:
            if value != default:
                return value
    return default


def get_env(name):
    value = os.environ.get(name)
    if value:
        return value
    else:
        LOG.debug("Environment variable %r is not defined", name)
        return None


def get_int_env(name):
    value = get_env(name)
    if value:
        try:
            return int(value)
        except TypeError:
            LOG.exception("Environment variable %r is not an integer: %r",
                          name, value)
    return None


def get_bool_env(name):
    value = get_env(name)
    if value:
        value = str(value).lower()
        if value in ['true', '1', 'yes']:
            return True
        elif value in ['false', '0', 'no']:
            return False
        else:
            LOG.exception("Environment variable %r is not a boolean: %r",
                          name, value)
    return None


def get_list_env(name, separator=','):
    value = get_env(name)
    if value:
        return value.split(separator)
    else:
        return []


def is_prevent_create() -> bool:
    return get_bool_env('TOBIKO_PREVENT_CREATE') is True


def skip_if_prevent_create(reason='TOBIKO_PREVENT_CREATE is True'):
    return tobiko.skip_if(reason=reason,
                          predicate=is_prevent_create)


def skip_unless_prevent_create(reason='TOBIKO_PREVENT_CREATE is False'):
    return tobiko.skip_unless(reason=reason,
                              predicate=is_prevent_create)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import logging

from oslo_log import log
from testtools import content

from tobiko.common import _detail
from tobiko.common import _fixture


LOG = log.getLogger(__name__)


class CaptureLogFixture(_fixture.SharedFixture):

    level = None
    logger = logging.root
    handler = None
    format = "%(asctime)-15s %(levelname)s %(name)s | %(message)s"

    def __init__(self, test_case_id, logger=None, level=None, fmt=None):
        super(CaptureLogFixture, self).__init__()
        self.test_case_id = test_case_id
        if logger:
            self.logger = logger
        if level:
            self.level = level
        if fmt:
            self.format = fmt

    def setup_fixture(self):
        self.handler = handler = CaptureLogHandler(level=self.level)
        formatter = logging.Formatter(self.format)
        handler.setFormatter(formatter)
        self.addCleanup(self.logger.removeHandler, handler)
        self.logger.addHandler(handler)
        LOG.debug('--- BEGIN %s ---', self.test_case_id)
        self.addCleanup(LOG.debug, '--- END %s ---', self.test_case_id)

    def getDetails(self):
        handler = self.handler
        if handler:
            content_object = _detail.details_content(
                content_type=content.UTF8_TEXT,
                content_id=self.fixture_name,
                get_text=handler.format_all)
            return {'log': content_object}
        else:
            return {}


class CaptureLogHandler(logging.Handler):

    def __init__(self, level=None):
        if level is None:
            from tobiko import config
            CONF = config.CONF
            if CONF.tobiko.debug:
                level = logging.DEBUG
            else:
                level = logging.INFO
        super(CaptureLogHandler, self).__init__(level)
        self.records = []

    def emit(self, record):
        self.records.append(record)

    def format_all(self):
        for record in self.records:
            yield self.format(record) + '\n'



# Copyright 2023 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools

from oslo_config import cfg


def setup_tobiko_config(conf):
    # pylint: disable=unused-argument
    from tobiko.podified import _openshift
    from tobiko.podified import _topology

    if _openshift.has_podified_cp():
        _topology.setup_podified_topology()


GROUP_NAME = "podified"
OPTIONS = [
    cfg.StrOpt('osp_project',
               default='openstack',
               help="Openshift project that includes the Openstack resources"),
]


def register_tobiko_options(conf):
    conf.register_opts(group=cfg.OptGroup(GROUP_NAME), opts=OPTIONS)


def list_options():
    return [(GROUP_NAME, itertools.chain(OPTIONS))]



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os

from oslo_log import log

import tobiko
from tobiko.shell import sh


LOG = log.getLogger(__name__)


class KeyFileFixture(tobiko.SharedFixture):

    def __init__(self,
                 key_file: str = None,
                 key_type: str = None):
        super(KeyFileFixture, self).__init__()
        if key_file is None:
            key_file = tobiko.tobiko_config().nova.key_file
        self.key_file = tobiko.tobiko_config_path(key_file)
        if key_type is None:
            key_type = tobiko.tobiko_config().nova.key_type
        self.key_type = key_type

    def setup_fixture(self):
        self.ensure_key_file()

    def ensure_key_file(self):
        key_file = tobiko.check_valid_type(self.key_file, str)
        LOG.debug(f'Ensuring Nova key files exist: {key_file}')
        if os.path.isfile(key_file):
            if os.path.isfile(f'{key_file}.pub'):
                LOG.info(f"Key file found: {key_file}")
                return
            else:
                LOG.info(f"Public key file not found: {key_file}.pub")
        else:
            LOG.info(f"Key file not found: {key_file}")

        LOG.info(f"Creating file for key pairs: '{self.key_file}'...")
        key_dir = os.path.dirname(key_file)
        tobiko.makedirs(key_dir)
        try:
            sh.local_execute(['ssh-keygen',
                              '-f', key_file,
                              '-P', '',
                              '-t', self.key_type])
        except sh.ShellCommandFailed:
            if (not os.path.isfile(key_file) or
                    not os.path.isfile(key_file + '.pub')):
                raise
        else:
            assert os.path.isfile(key_file)
            assert os.path.isfile(key_file + '.pub')
            LOG.info(f'Key file created: {self.key_file}')


def get_key_file() -> str:
    return tobiko.setup_fixture(KeyFileFixture).key_file



"""Error classes and wrappers for VarlinkError."""

from __future__ import absolute_import

from varlink import VarlinkError


class VarlinkErrorProxy(VarlinkError):
    """Class to Proxy VarlinkError methods."""

    def __init__(self, message, namespaced=False):
        """Construct proxy from Exception."""
        super().__init__(message.as_dict(), namespaced)
        self._message = message
        self.__module__ = 'libpod'

    def __getattr__(self, method):
        """Return attribute from proxied Exception."""
        if hasattr(self._message, method):
            return getattr(self._message, method)

        try:
            return self._message.parameters()[method]
        except KeyError as ex:
            raise AttributeError(
                'No such attribute: {}'.format(method)) from ex


class ContainerNotFound(VarlinkErrorProxy):
    """Raised when Client cannot find requested container."""


class ImageNotFound(VarlinkErrorProxy):
    """Raised when Client cannot find requested image."""


class PodNotFound(VarlinkErrorProxy):
    """Raised when Client cannot find requested image."""


class PodContainerError(VarlinkErrorProxy):
    """Raised when a container fails requested pod operation."""


class NoContainerRunning(VarlinkErrorProxy):
    """Raised when no container is running in pod."""


class NoContainersInPod(VarlinkErrorProxy):
    """Raised when Client fails to connect to runtime."""


class ErrorOccurred(VarlinkErrorProxy):
    """Raised when an error occurs during the execution.

    See error() to see actual error text.
    """


class PodmanError(VarlinkErrorProxy):
    """Raised when Client fails to connect to runtime."""


class InvalidState(VarlinkErrorProxy):
    """Raised when container is in invalid state for operation."""


ERROR_MAP = {
    'io.podman.ContainerNotFound': ContainerNotFound,
    'io.podman.ErrorOccurred': ErrorOccurred,
    'io.podman.ImageNotFound': ImageNotFound,
    'io.podman.InvalidState': InvalidState,
    'io.podman.NoContainerRunning': NoContainerRunning,
    'io.podman.NoContainersInPod': NoContainersInPod,
    'io.podman.PodContainerError': PodContainerError,
    'io.podman.PodNotFound': PodNotFound,
    'io.podman.RuntimeError': PodmanError,
}


def error_factory(exception):
    """Map Exceptions to a discrete type."""
    try:
        return ERROR_MAP[exception.error()](exception)
    except KeyError:
        return exception



from __future__ import absolute_import

import io
import re

from oslo_log import log
import pandas

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack import topology
from tobiko.tripleo import overcloud
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


class OvercloudProcessesException(tobiko.TobikoException):
    message = "not all overcloud processes are in running state, " \
              "{process_error}"


def get_overcloud_node_processes_table(ssh_client: ssh.SSHClientType):
    """
    get processes tables from overcloud node

       returns :
[root@controller-0 ~]# ps -axw -o "%U" -o "|%p" -o "|%P" -o "|%C" -o "|%z" -o
"|%x" -o "|%c" -o "|%a" |grep -v 'ps -aux'|head
USER    |    PID|   PPID|%CPU|   VSZ|    TIME|COMMAND        |COMMAND
|overcloud_node
root    |      1|      0| 1.3|246892|01:08:57|systemd        |/usr/lib/systemd
controller-0 ...
/systemd --switched-root --system --deserialize 18
root    |      2|      0| 0.0|     0|00:00:00|kthreadd       |[kthreadd]
root    |      3|      2| 0.0|     0|00:00:00|rcu_gp         |[rcu_gp]
root    |      4|      2| 0.0|     0|00:00:00|rcu_par_gp     |[rcu_par_gp]
root    |      6|      2| 0.0|     0|00:00:00|kworker/0:0H-ev|[kworker/0:0H
-events_highpri]
root    |      8|      2| 0.0|     0|00:00:00|mm_percpu_wq   |[mm_percpu_wq]
root    |      9|      2| 0.0|     0|00:00:06|ksoftirqd/0    |[ksoftirqd/0]
root    |     10|      2| 0.0|     0|00:04:28|rcu_sched      |[rcu_sched]
root    |     11|      2| 0.0|     0|00:00:05|migration/0    |[migration/0]

    :return: dataframe of overcloud node processes dataframe
    """

    output = sh.execute(
        "ps -axw -o \"%U\" -o \"DELIM%p\" -o \"DELIM%P\" -o \"DELIM%C\" -o "
        "\"DELIM%z\" -o \"DELIM%x\" -o \"DELIM%c\" -o \"DELIM%a\" |grep -v "
        "'ps -axw' |sed 's/\"/''/g'",
        ssh_client=ssh_client).stdout
    stream = io.StringIO(output)
    table: pandas.DataFrame = pandas.read_csv(
        stream, sep='DELIM', header=None, skiprows=1)
    table.replace(to_replace=' ', value="", regex=True, inplace=True)
    table.columns = ['USER', 'PID', 'PPID', 'CPU', 'VSZ', 'TIME', 'PROCESS',
                     'PROCESS_ARGS']
    # pylint: disable=unsupported-assignment-operation
    hostname = sh.get_hostname(ssh_client=ssh_client)
    table['overcloud_node'] = hostname

    LOG.debug("Successfully got overcloud nodes processes status table")
    return table


def get_overcloud_nodes_running_process(process):
    """
    Check what nodes are running the specifies
    process: exact str of a process name as seen in ps -axw -o "%c"
    :return: list of overcloud nodes
    """
    oc_procs_df = overcloud.get_overcloud_nodes_dataframe(
                                            get_overcloud_node_processes_table)
    oc_nodes_running_process = oc_procs_df.query('PROCESS=="{}"'.format(
        process))['overcloud_node'].unique()
    return oc_nodes_running_process


def check_if_process_running_on_overcloud(process):
    """
    Check what nodes are running the specifies
    process: exact str of a process name as seen in ps -axw -o "%c"
    :return: list of overcloud nodes
    """
    oc_procs_df = overcloud.get_overcloud_nodes_dataframe(
                                          get_overcloud_node_processes_table)
    if not oc_procs_df.query('PROCESS=="{}"'.format(process)).empty:
        return True
    else:
        return False


class OvercloudProcessesStatus(object):
    """
    class to handle processes checks,
    checks that all of these are running in the overcloud:
    'ovsdb-server','pcsd', 'corosync', 'beam.smp', 'mysqld', 'redis-server',
    'haproxy', 'nova-conductor', 'nova-scheduler', 'neutron-server',
    'nova-compute', 'glance-api'
    """
    def __init__(self):
        self.processes_to_check = ['ovsdb-server', 'pcsd', 'corosync',
                                   'beam.smp', 'mysqld', 'redis-server',
                                   'haproxy', 'nova-conductor',
                                   'nova-scheduler', 'neutron-server',
                                   'nova-compute', 'glance-api']

        num_northd_proc = 'all' if overcloud.is_ovn_using_raft() else 1
        self.ovn_processes_to_check_per_node = [{'name': 'ovn-controller',
                                                 'node_group': 'controller',
                                                 'number': 'all'},
                                                {'name': 'ovn-controller',
                                                 'node_group': 'compute',
                                                 'number': 'all'},
                                                {'name': 'ovn-northd',
                                                 'node_group': 'controller',
                                                 'number': num_northd_proc}]

        self.oc_procs_df = overcloud.get_overcloud_nodes_dataframe(
                                            get_overcloud_node_processes_table)

    def _basic_overcloud_process_running(self, process_name):
        # osp16/python3 process is "neutron-server:"
        if process_name == 'neutron-server' and \
                self.oc_procs_df.query('PROCESS=="{}"'.format(
                process_name)).empty:
            process_name = 'neutron-server:'
        # osp17 mysqld process name is mysqld_safe
        if process_name == 'mysqld' and \
                self.oc_procs_df.query('PROCESS=="{}"'.format(
                process_name)).empty:
            process_name = 'mysqld_safe'
        # redis not deployed on osp17 by default, only if some
        # other services such as designate and octavia are deployed
        if (process_name == 'redis-server' and
                not overcloud.is_redis_expected()):
            redis_message = ("redis-server not expected on OSP 17 "
                             "and later releases by default")
            if self.oc_procs_df.query(
                    f'PROCESS=="{process_name}"').empty:
                LOG.info(redis_message)
                return
            else:
                raise OvercloudProcessesException(
                    process_error=redis_message)

        if not self.oc_procs_df.query('PROCESS=="{}"'.format(
                process_name)).empty:
            LOG.info("overcloud processes status checks: "
                     "process {} is  "
                     "in running state".format(process_name))
            return
        else:
            LOG.info("Failure : overcloud processes status checks:"
                     "process {} is not running ".format(
                      process_name))
            raise OvercloudProcessesException(
                process_error="process {} is not running ".format(
                              process_name))

    @property
    def basic_overcloud_processes_running(self):
        """
        Checks that the oc_procs_df dataframe has all of the list procs
        :return: Bool
        """
        for attempt in tobiko.retry(timeout=300., interval=1.):
            try:
                for process_name in self.processes_to_check:
                    self._basic_overcloud_process_running(process_name)
            except OvercloudProcessesException:
                if attempt.is_last:
                    LOG.error('Not all overcloud processes are running')
                    raise
                LOG.info('Retrying overcloud processes: %s', attempt.details)
                self.oc_procs_df = overcloud.get_overcloud_nodes_dataframe(
                    get_overcloud_node_processes_table)

            # if all procs are running we can return true
            return True

    def _ovn_overcloud_process_validations(self, process_dict):
        if not self.oc_procs_df.query('PROCESS=="{}"'.format(
                process_dict['name'])).empty:
            LOG.info("overcloud processes status checks: "
                     f"process {process_dict['name']} is  "
                     "in running state")

            ovn_proc_filtered_df = self.oc_procs_df.query(
                'PROCESS=="{}"'.format(process_dict['name']))

            if (process_dict['node_group'] not in
                    topology.list_openstack_node_groups()):
                LOG.debug(
                    f"{process_dict['node_group']} is not "
                    "a node group part of this Openstack cloud")
                return
            node_list = [node.name
                         for node in
                         topology.list_openstack_nodes(
                            group=process_dict['node_group'])]
            node_names_re = re.compile(r'|'.join(node_list))
            node_filter = (ovn_proc_filtered_df.overcloud_node.
                           str.match(node_names_re))
            # get the processes running on a specific type of nodes
            ovn_proc_filtered_per_node_df = \
                ovn_proc_filtered_df[node_filter]
            total_num_processes = len(ovn_proc_filtered_per_node_df)

            if type(process_dict['number']) == int:
                expected_num_processes = process_dict['number']
            elif process_dict['number'] == 'all':
                expected_num_processes = len(node_list)
            else:
                raise ValueError("Unexpected value:"
                                 f"{process_dict['number']}")

            if expected_num_processes != total_num_processes:
                raise OvercloudProcessesException(
                    "Unexpected number"
                    f" of processes {process_dict['name']} running on "
                    f"{process_dict['node_group']} nodes")
            # process successfully validated
            LOG.debug(f"{process_dict['name']} successfully validated on "
                      f"{process_dict['node_group']} nodes")

    @property
    def ovn_overcloud_processes_validations(self):
        """
        Checks that the oc_procs_df dataframe has OVN processes running on the
        expected overcloud node or nodes
        :return: Bool
        """
        if not neutron.has_ovn():
            LOG.info("Networking OVN not configured")
            return True

        for attempt in tobiko.retry(timeout=300., interval=1.):
            try:
                for process_dict in self.ovn_processes_to_check_per_node:
                    self._ovn_overcloud_process_validations(process_dict)
            except OvercloudProcessesException:
                if attempt.is_last:
                    LOG.error('Unexpected number of OVN overcloud processes')
                    raise
                LOG.info('Retrying OVN overcloud processes: %s',
                         attempt.details)
                self.oc_procs_df = overcloud.get_overcloud_nodes_dataframe(
                    get_overcloud_node_processes_table)

            # if all procs are running we can return true
            return True



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing
import os

import tobiko


class RunConfigFixture(tobiko.SharedFixture):

    test_path: typing.List[str]
    test_filename: str = 'test_*.py'
    python_path: typing.Optional[typing.List[str]] = None
    workers_count: typing.Optional[int] = None

    def setup_fixture(self):
        package_file = os.path.realpath(os.path.realpath(tobiko.__file__))
        package_dir = os.path.dirname(package_file)
        tobiko_dir = os.path.dirname(package_dir)
        self.test_path = [os.path.join(tobiko_dir, 'tobiko', 'tests', 'unit')]

    @property
    def forked(self) -> bool:
        return self.workers_count is not None and self.workers_count != 1


def run_confing(obj=None) -> RunConfigFixture:
    if obj is None:
        return tobiko.setup_fixture(RunConfigFixture)
    fixture = tobiko.get_fixture(obj)
    tobiko.check_valid_type(fixture, RunConfigFixture)
    return tobiko.setup_fixture(fixture)



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from heatclient.v1 import resources

import tobiko
from tobiko.openstack.heat import _client
from tobiko.openstack.heat import _stack


RESOURCE_CLASSES = resources.Resource,
ResourceType = typing.Union[resources.Resource]


def list_resources(stack: _stack.StackIdType,
                   client: _client.HeatClientType = None,
                   **kwargs) -> tobiko.Selection[ResourceType]:
    client = _client.heat_client(client)
    stack_id = _stack.get_stack_id(stack)
    return tobiko.select(client.resources.list(stack_id=stack_id, **kwargs))


def find_resource(stack: _stack.StackIdType,
                  client: _client.HeatClientType = None,
                  unique=False,
                  **kwargs) -> ResourceType:
    _stacks = list_resources(stack=stack, client=client, **kwargs)
    if unique:
        return _stacks.unique
    else:
        return _stacks.first
